# -*- coding: utf-8 -*-
"""
This calss contains the AFGL Atmospheric Constituent Profiles (0-120km).
"""
import os
from typing import Tuple, Dict
import numpy as np


class AtmosphericProfiles:
    """AFGL Atmospheric Constituent Profiles (0-120km)

    Each of these profile contains data at 50 atmospheric levels.
    Altitude (km), Pressure (mb), Density (cm-3), Molec. densities (ppmv):
    
    * 0 (H2O),
    * 1 (CO2),
    * 2 (O3),
    * 3 (N2O),
    * 4 (CO),
    * 5 (CH4),
    * 6 (O2)
    
    Plus suplimental profiles where available.
    The last set of data sets are constituent profiles of molecular
    densities (ppmv) for the minor absorbing atmospheric gases.

    References
    ----------

    .. [1] [ANDERSON]_
    
    Examples:
        .. code-block:: python
        
            >>> from pyrtlib.climatology import AtmosphericProfiles as atmp
            >>> atmp.atm_profiles()
            {0: 'Tropical',
            1: 'Midlatitude Summer',
            2: 'Midlatitude Winter',
            3: 'Subarctic Summer',
            4: 'Subarctic Winter',
            5: 'US Standard'}
            >>> atmp.TROPICAL, atmp.H2O
            (0, 0)
    """

    TROPICAL = 0
    MIDLATITUDE_SUMMER = 1
    MIDLATITUDE_WINTER = 2
    SUBARCTIC_SUMMER = 3
    SUBARCTIC_WINTER = 4
    US_STANDARD = 5

    H2O = 0
    CO2 = 1
    O3 = 2
    N2O = 3
    CO = 4
    CH4 = 5
    O2 = 6
    # MINOR
    NO = 7
    SO2 = 8
    NO2 = 9
    NH3 = 10
    HNO3 = 11
    OH = 12
    HF = 13
    HCL = 14
    HBR = 15
    HI = 16
    CLO = 17
    OCS = 18
    H2CO = 19
    HOCL = 20
    N2 = 21
    HCN = 22
    CH3CL = 23
    H2O2 = 24
    C2H2 = 25
    C2H6 = 26
    PH3 = 27
    # TRACE
    COF2 = 28
    SF6 = 29
    H2S = 30
    CCL3F = 31
    CCL2F2 = 32
    CCLF3 = 33
    CF4 = 34
    CHCl2F = 35
    CHCLF2 = 36
    C2CL3F3 = 37
    C2CL2F4 = 38
    C2CLF5 = 39
    CCL4 = 40
    CLONO2 = 41
    N2O5 = 42
    HNO4 = 43
    AIR = 99

    @staticmethod
    def atm_profiles() -> Dict[int, str]:
        """Convenient function to ger the list of the buolt-in  Atmospheric profiles.

        Returns:
            Dict[int, str]: A dictionary of standard profiles atmospheric.
        """
        return {AtmosphericProfiles.TROPICAL: 'Tropical',
                AtmosphericProfiles.MIDLATITUDE_SUMMER: 'Midlatitude Summer',
                AtmosphericProfiles.MIDLATITUDE_WINTER: 'Midlatitude Winter',
                AtmosphericProfiles.SUBARCTIC_SUMMER: 'Subarctic Summer',
                AtmosphericProfiles.SUBARCTIC_WINTER: 'Subarctic Winter',
                AtmosphericProfiles.US_STANDARD: 'US Standard'}


    @staticmethod
    def gl_atm(atm: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Returns the Atmopshere profile.
        
        This method contains 6 model profiles:
        
        +--------+---------------------+
        | option | model               |
        +--------+---------------------+
        | 1      |  Tropical           |
        +--------+---------------------+
        | 2      |  Midlatitude Summer |
        +--------+---------------------+
        | 3      |  Midlatitude Winter |
        +--------+---------------------+
        | 4      |  Subarctic Summer   |
        +--------+---------------------+
        | 5      |  Subarctic Winter   |
        +--------+---------------------+
        | 6      |  U.S. Standard      |
        +--------+---------------------+

        Args:
            option (int): the atmosphere profile.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.ndarray]: 
            * a (numpy.ndarray): Altitudes (km) (50x1)
            * p (numpy.ndarray): Pressure (mbar)
            * d (numpy.ndarray): Total density (cm-3)
            * t (numpy.ndarray): Temperature (K)
            * md (numpy.ndarray): Molecular densities (ppmv)

        Example:
            >>> from pyrtlib.climatology import AtmosphericProfiles as atmp
            >>> z, p, _, t, md = atmp.gl_atm(atmp.US_STANDARD)
            >>> md[:, atmp.H2O]
            array([7.745e+03, 6.071e+03, 4.631e+03, 3.182e+03, 2.158e+03, 1.397e+03,
                  9.254e+02, 5.720e+02, 3.667e+02, 1.583e+02, 6.996e+01, 3.613e+01,
                  1.906e+01, 1.085e+01, 5.927e+00, 5.000e+00, 3.950e+00, 3.850e+00,
                  3.825e+00, 3.850e+00, 3.900e+00, 3.975e+00, 4.065e+00, 4.200e+00,
                  4.300e+00, 4.425e+00, 4.575e+00, 4.725e+00, 4.825e+00, 4.900e+00,
                  4.950e+00, 5.025e+00, 5.150e+00, 5.225e+00, 5.250e+00, 5.225e+00,
                  5.100e+00, 4.750e+00, 4.200e+00, 3.500e+00, 2.825e+00, 2.050e+00,
                  1.330e+00, 8.500e-01, 5.400e-01, 4.000e-01, 3.400e-01, 2.800e-01,
                  2.400e-01, 2.000e-01])

        .. note:: adapted from glatm.dat.  DCT 3/26/97
        """

        path = os.path.dirname(os.path.abspath(__file__))
        option = atm + 1
        # MODEL 1.  TROPICAL
        if option == 1:
            #
            # Latitude (deg)
            # 15.0
            mtx = np.loadtxt(os.path.join(path, "tropical.dat"))

        # MODEL 2. MIDLATITUDE SUMMER
        elif option == 2:
            #
            # Latitude (deg)
            # 45.0
            mtx = np.loadtxt(os.path.join(path, "midlatitude_summer.dat"))

        # MODEL 3. MIDLATITUDE WINTER
        elif option == 3:
            #
            # Latitude (deg)
            # 45.0
            mtx = np.loadtxt(os.path.join(path, "midlatitude_winter.dat"))

        # MODEL 4. SUBARCTIC SUMMER
        elif option == 4:
            #
            # Latitude (deg)
            # 60.0
            #
            mtx = np.loadtxt(os.path.join(path, "subarctic_summer.dat"))

        # MODEL 5. SUBARCTIC WINTER
        elif option == 5:
            #
            # Latitude (deg)
            # 60.0
            mtx = np.loadtxt(os.path.join(path, "subarctic_winter.dat"))

        # MODEL 6. U.S. STANDARD
        elif option == 6:
            #
            # Latitude (deg)
            # 45.5397
            mtx = np.loadtxt(os.path.join(path, "us_standard.dat"))
        
        # Altitude (km)
        a = mtx[:, 0]
        # Pressue (mb)
        p = mtx[:, 1]
        # Temperature (K)
        t = mtx[:, 3]
        # Density (cm-3)
        d = mtx[:, 2]
        # 1 :H2O (ppmv)
        g1 = mtx[:, 4]
        # 2 :CO2 (ppmv)
        g2 = mtx[:, 5]
        # 3 :O3 (ppmv)
        g3 = mtx[:, 6]
        # 4 : N2O (ppmv)
        g4 = mtx[:, 7]
        # 5 :CO (ppmv)
        g5 = mtx[:, 8]
        # 6 :CH4 (ppmv)
        g6 = mtx[:, 9]
        # 7 :O2 (ppmv)
        g7 = mtx[:, 10]

        md = np.column_stack((g1, g2, g3, g4, g5, g6, g7))

        return a, p, d, t, md

    @staticmethod
    def gl_atm_minor(gas_minor: int) -> np.ndarray:
        """Returns the minor gas profiles (gas ID's 8-28)

        Args:
            gas_minor (int): HITRAN gas ID #'s (#gases x 1)

        Returns:
            numpy.ndarray: molecular densities (ppmv) (50x #gases)

        Example:

            >>> from pyrtlib.climatology import AtmosphericProfiles as atmp
            >>> atmp.gl_atm_minor(atmp.NO)
            array([3.00e-04, 3.00e-04, 3.00e-04, 3.00e-04, 3.00e-04, 3.00e-04,
                   3.00e-04, 3.00e-04, 3.00e-04, 3.00e-04, 3.00e-04, 3.00e-04,
                   3.00e-04, 2.99e-04, 2.95e-04, 2.83e-04, 2.68e-04, 2.52e-04,
                   2.40e-04, 2.44e-04, 2.55e-04, 2.77e-04, 3.07e-04, 3.60e-04,
                   4.51e-04, 6.85e-04, 1.28e-03, 2.45e-03, 4.53e-03, 7.14e-03,
                   9.34e-03, 1.12e-02, 1.19e-02, 1.17e-02, 1.10e-02, 1.03e-02,
                   1.01e-02, 1.01e-02, 1.03e-02, 1.15e-02, 1.61e-02, 2.68e-02,
                   7.01e-02, 2.13e-01, 7.12e-01, 2.08e+00, 4.50e+00, 7.98e+00,
                   1.00e+01, 1.00e+01])
        """

        path = os.path.dirname(os.path.abspath(__file__))
        mtx = np.loadtxt(os.path.join(path, "gas_minor.dat"))
        gas_minor -= 7
        # CONSTITUENT PROFILES FOR THE MINOR ABSORBING ATMOSPHERIC GASES
        #
        # MINGAS
        g = mtx[:, gas_minor]

        return g

    @staticmethod
    def gl_atm_trace(gas_trace: int) -> np.ndarray:
        """Returns the trace gas profiles (ID's 29-31,51-63)

        Args:
            gas_trace (int): HITRAN gas ID

        Returns:
            numpy.ndarray: molecular densities (ppmv) (50x #gases)

        Example:

            >>> from pyrtlib.climatology import AtmosphericProfiles as atmp
            >>> atmp.gl_atm_trace(atmp.H2S)
            array([1.00e-04, 5.48e-05, 3.00e-05, 2.45e-05, 2.00e-05, 1.61e-05,
                  1.30e-05, 1.14e-05, 1.00e-05, 8.37e-06, 7.00e-06, 4.58e-06,
                  3.00e-06, 1.73e-06, 1.00e-06, 5.48e-07, 3.00e-07, 1.73e-07,
                  1.00e-07, 5.48e-08, 3.00e-08, 1.73e-08, 1.00e-08, 5.48e-09,
                  3.00e-09, 1.73e-09, 1.00e-09, 1.00e-10, 1.00e-11, 1.00e-12,
                  1.00e-13, 1.00e-14, 1.00e-15, 1.00e-15, 1.00e-15, 1.00e-15,
                  1.00e-15, 1.00e-15, 1.00e-15, 1.00e-15, 1.00e-15, 1.00e-15,
                  1.00e-15, 1.00e-15, 1.00e-15, 1.00e-15, 1.00e-15, 1.00e-15,
                  1.00e-15, 1.00e-15])
        
        References
        ----------

            .. [1] Smith, M.A.H., Compilation of atmospheric gas concentration profiles from 0 to 50km, NASA Tech.Mem. 83289, 1982
            .. [2] Rinsland, C.P., et al., JGR 94 D15 18341-18349, 1989
            .. [3] Rinsland, C.P., et al., JGR 95 D10 16477-16490, 1990
        
        """

        # ************************************************************************
        # ATMOSPHERIC PROFILES OF THE MINOR GASES (ID>28)
        # REFERENCES
        # ATMOS with no further reference refers to the profile
        # set distributed by diskette
        # Smith, M.A.H., Compilation of atmospheric gas
        # concentration profiles from 0 to 50km,
        # NASA Tech.Mem. 83289, 1982
        # Rinsland, C.P., et al., JGR 94 D15 18341-18349, 1989
        #
        # Rinsland, C.P., et al., JGR 95 D10 16477-16490, 1990
        # Copy of FASCODE XMLATM BLOCK DATA
        # Data from Mark Allen at JPL/private communication to Gail Anderson
        # ************************************************************************

        path = os.path.dirname(os.path.abspath(__file__))
        mtx = np.loadtxt(os.path.join(path, "gas_trace.dat"))
        gas_trace -= 28
        #  COF2 & SF6
        #   0 - 14   ATMOS SS Rinsland 1990: ConsAt 14 value
        #  14 -120   ATMOS SS Rinsland 1990: LinExLogMix until 1.0E-15
        
        #  H2S, CCLF3  (CFC13), CHCl2F (CFC21), C2CLF5  (CFC115)
        #   0 - 120   Smith 1982: until 1.0E-15
        #  CCL3F (F11), CCL2F2 (CFC12), CHCLF2 (CFC22), C2CL3F3 (CFC113), C2CL2F4 (CFC114), CCL4, CLONO2, N2O5, HNO4
        #   0 -120   Mark Allen/FASCODE
        #  CF4 (CFC14)
        #   0 - 10   ATMOS SS: ConsAt 10.5 value
        #  11 - 30   ATMOS SS: LinExLogMix
        #  30 - 75   ATMOS SS: ConsAt 31.5 value
        #  75 -120   1.0E-15
        g = mtx[:, gas_trace]

        return g



"""Utility code to support making requests using HTTP.
The orignal code of this class can be found in https://github.com/Unidata/siphon
Terms and conditions are as for siphon library license (https://github.com/Unidata/siphon/blob/master/LICENSE)
"""

import posixpath
from io import BytesIO
from typing import Optional
from urllib.parse import urlencode, urljoin  # noqa

import requests
from requests.sessions import Session

from ..version import __version__

HTTPError = requests.HTTPError


class HTTPSessionManager(object):
    """Manage the creation of sessions for HTTP access."""

    def __init__(self):
        """Initialize ``HTTPSessionManager``."""
        self.user_agent = 'pyrtlib ({})'.format(__version__)
        self.options = {}

    def set_session_options(self, **kwargs) -> None:
        """Set options for created session instances.

        Takes keyword arguments and sets them as attributes on the returned
        :class:`requests.Session` instance.

        See Also
        --------
        create_session

        """
        self.options = kwargs

    def create_session(self) -> Session:
        """Create a new HTTP session with our user-agent set.

        Returns
        -------
        session : requests.Session
            The created session

        See Also
        --------
        urlopen, set_session_options

        """
        ret = requests.Session()
        ret.headers['User-Agent'] = self.user_agent
        for k, v in self.options.items():
            setattr(ret, k, v)
        return ret

    def urlopen(self, url: str, **kwargs) -> BytesIO:
        """GET a file-like object for a URL using HTTP.

        This is a thin wrapper around :meth:`requests.Session.get` that returns a file-like
        object wrapped around the resulting content.

        Parameters
        ----------
        url : str
            The URL to request

        kwargs : arbitrary keyword arguments
            Additional keyword arguments to pass to :meth:`requests.Session.get`.

        Returns
        -------
        fobj : file-like object
            A file-like interface to the content in the response

        See Also
        --------
        :meth:`requests.Session.get`

        """
        return BytesIO(self.create_session().get(url, **kwargs).content)


session_manager = HTTPSessionManager()


class HTTPEndPoint(object):
    """An object representing an endpoint on a server that is accessed using HTTP.

    This provides a simple way to point to a URL, formulate appropriate queries and
    validate them, parse metadata as appropriate, and parse returns from requests.
    """

    def __init__(self, url: str) -> None:
        """Create an HTTPEndPoint instance.

        Parameters
        ----------
        url : str
            The base URL for the endpoint

        """
        self._base = url
        self._session = session_manager.create_session()
        self._get_metadata()

    def _get_query(self, query: dict) -> str:
        """Make a GET request, including a query, to the endpoint.

        The path of the request is to the base URL assigned to the endpoint.

        Parameters
        ----------
        query : dict
            The query to pass when making the request

        Returns
        -------
        resp : requests.Response
            The server's response to the request

        See Also
        --------
        get_path, get

        """
        url = self._base[:-1] if self._base[-1] == '/' else self._base
        return self._get(url, query)

    def _url_path(self, path: str) -> str:
        """Assemble the full url to a path.

        Given a path relative to the base URL, assemble the full URL.

        Parameters
        ----------
        path : str
            The path, relative to the endpoint

        Returns
        -------
        url : str
            The full URL to `path`

        See Also
        --------
        get_path

        """
        return posixpath.join(self._base, path)

    def _get_path(self, path: str, querystrings: Optional[dict] = None) -> requests.Response:
        """Make a GET request, optionally including a query, to a relative path.

        The path of the request includes a path on top of the base URL
        assigned to the endpoint.

        Parameters
        ----------
        path : str
            The path to request, relative to the endpoint
        querystrings : dict, optional
            The query to pass when making the request

        Returns
        -------
        resp : requests.Response
            The server's response to the request

        See Also
        --------
        get_query, get, url_path

        """
        return self._get(self._url_path(path), querystrings)

    def _get(self, path: str, params: Optional[dict] = None) -> str:
        """Make a GET request, optionally including a parameters, to a path.

        The path of the request is the full URL.

        Parameters
        ----------
        path : str
            The URL to request
        params : dict, optional
            The query to pass when making the request

        Returns
        -------
        resp : requests.Response
            The server's response to the request

        Raises
        ------
        HTTPError
            If the server returns anything other than a 200 (OK) code

        See Also
        --------
        get_query, get

        """
        resp = self._session.get(path, params=params)
        if resp.status_code != 200:
            if resp.headers.get('Content-Type', '').startswith('text/html'):
                text = resp.reason
            else:
                text = resp.text
            raise requests.HTTPError('Error accessing {0}\n'
                                     'Server Error ({1:d}: {2})'.format(resp.request.url,
                                                                        resp.status_code,
                                                                        text))
        return resp

    def _get_metadata(self):
        """Get the metadata associated with the endpoint.

        It is intended that this be implemented by subclasses as necessary.
        """



import numpy as np
import os

PATH = os.path.dirname(os.path.abspath(__file__))
os.path.join(PATH, "o3_list_r22.txt")
mtx = np.loadtxt(os.path.join(PATH, "o3_list_r22.txt"))

reftline = 296
# line intensities from hitran include isotopomer abundance
# read line parameters; units: ghz, hz*cm^2, mhz/mb
fl = mtx[:, 1]
s1 = mtx[:, 2]
b = mtx[:, 3]
w = mtx[:, 4] / 1000.0
x = mtx[:, 5]
sr = mtx[:, 6]



"""Read upper air data from the Integrated Global Radiosonde Archive version 2.
The orignal code of this class can be found in https://github.com/Unidata/siphon
Terms and conditions are as for siphon library license (https://github.com/Unidata/siphon/blob/master/LICENSE)
"""

__author__ = ''
__date__ = 'March 2021'
__copyright__ = '(C) 2021, CNR-IMAA'

from datetime import datetime
from io import BytesIO
from io import StringIO
import itertools
import sys
import warnings
from zipfile import ZipFile
from typing import List, Union, Optional, Tuple, Dict

import numpy as np
import pandas as pd

from ..apiwebservices.webservices import HTTPEndPoint, HTTPError

warnings.filterwarnings('ignore', "Pandas doesn't allow columns to be created", UserWarning)

def get_wind_components(speed, wdir):
    r"""Calculate the U, V wind vector components from the speed and direction.
    Parameters
    ----------
    speed : array_like
        The wind speed (magnitude)
    wdir : array_like
        The wind direction, specified as the direction from which the wind is
        blowing, with 0 being North.
    Returns
    -------
    u, v : tuple of array_like
        The wind components in the X (East-West) and Y (North-South)
        directions, respectively.
    """
    u = -speed * np.sin(wdir)
    v = -speed * np.cos(wdir)
    return u, v


class IGRAUpperAir(HTTPEndPoint):
    """Download and parse data from NCEI's Integrated Radiosonde Archive version 2."""

    def __init__(self) -> None:
        """Set http site address and file suffix based on desired dataset."""
        self.suffix = ''
        self.begin_date = ''
        self.end_date = ''
        self.site_id = ''
        self.folder = ''
        super().__init__('https://www1.ncdc.noaa.gov/pub/data/igra/')

    @classmethod
    def request_data(cls, time: datetime, site_id: str, beg2021: Optional[bool] = False, derived: Optional[bool] = False) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """Retreive IGRA version 2 data files contain the full period of record.

        Args:
            time (datetime): The date and time of the desired observation. If list of two times is given,
                dataframes for all dates within the two dates will be returned.
            site_id (str): 11-character IGRA2 station identifier.
            beg2021 (Optional[bool], optional): If True retrieve files of derived sounding parameters. Defaults to False.
            derived (Optional[bool], optional): If True retrieve files only contain soundings from the current
                (or current and previous) year. Defaults to False.

        Returns:
            Tuple[pandas.DataFrame, pandas.DataFrame]: A dataframe containing the data.
        """

        igra2 = cls()

        # Set parameters for data query
        if beg2021:
            igra2.folder = 'data/data-y2d/'
            igra2.suffix = igra2.suffix + '-data-beg2021.txt'
        else:
            igra2.folder = 'data/data-por/'
            igra2.suffix = igra2.suffix + '-data.txt'
            
        if derived:
            igra2.folder = 'derived/derived-por/'
            igra2.suffix = igra2.suffix + '-drvd.txt'

        if type(time) == datetime:
            igra2.begin_date = time
            igra2.end_date = time
        else:
            igra2.begin_date, igra2.end_date = time

        igra2.site_id = site_id

        df, headers = igra2._get_data()

        return df, headers

    def _get_data(self) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """Process the IGRA2 text file for observations at site_id matching time.

        Returns
        -------
            :class: `pandas.DataFrame` containing the body data.
            :class: `pandas.DataFrame` containing the header data.

        """
        # Split the list of times into begin and end dates. If only
        # one date is supplied, set both begin and end dates equal to that date.

        body, header, dates_long, dates = self._get_data_raw()

        params = self._get_fwf_params()

        df_body = pd.read_fwf(StringIO(body), **params['body'])
        df_header = pd.read_fwf(StringIO(header), **params['header'])
        df_body['date'] = dates_long

        df_body = self._clean_body_df(df_body)
        df_header = self._clean_header_df(df_header)
        df_header['date'] = dates

        return df_body, df_header

    def _get_data_raw(self) -> Tuple:
        """Download observations matching the time range.

        Returns a tuple with a string for the body, string for the headers,
        and a list of dates.
        """
        path = self.folder + self.site_id + self.suffix + '.zip'

        # Get the data and handle if there is none matching what was requested
        try:
            resp = self._get_path(path)
        except HTTPError:
            raise ValueError('No data available for {time:%Y-%m-%d %HZ} '
                             'for station {stid}.'.format(time=self.begin_date,
                                                          stid=self.site_id))

        file_info = ZipFile(BytesIO(resp.content)).infolist()[0]
        f = ZipFile(BytesIO(resp.content)).open(file_info)

        lines = [line.decode('utf-8') for line in f.readlines()]

        body, header, dates_long, dates = self._select_date_range(lines)

        return body, header, dates_long, dates

    def _select_date_range(self, lines: List) -> Tuple:
        """Identify lines containing headers within the range begin_date to end_date.

        Parameters
        ----------
        lines: list
            list of lines from the IGRA2 data file.

        """
        headers = []
        num_lev = []
        dates = []

        # Get indices of headers, and make a list of dates and num_lev
        for idx, line in enumerate(lines):
            if line[0] == '#':
                year, month, day, hour = map(int, line[13:26].split())

                # All soundings have YMD, most have hour
                try:
                    date = datetime(year, month, day, hour)
                except ValueError:
                    date = datetime(year, month, day)

                # Check date
                if self.begin_date <= date <= self.end_date:
                    headers.append(idx)
                    num_lev.append(int(line[32:36]))
                    dates.append(date)
                if date > self.end_date:
                    break

        if len(dates) == 0:
            # Break if no matched dates.
            # Could improve this later by showing the date range for the station.
            raise ValueError('No dates match selection.')

        # Compress body of data into a string
        begin_idx = min(headers)
        end_idx = max(headers) + num_lev[-1]

        # Make a boolean vector that selects only list indices within the time range
        selector = np.zeros(len(lines), dtype=bool)
        selector[begin_idx:end_idx + 1] = True
        selector[headers] = False
        body = ''.join(itertools.compress(lines, selector))

        selector[begin_idx:end_idx + 1] = ~selector[begin_idx:end_idx + 1]
        header = ''.join(itertools.compress(lines, selector))

        # expand date vector to match length of the body dataframe.
        dates_long = np.repeat(dates, num_lev)

        return body, header, dates_long, dates

    def _get_fwf_params(self) -> Dict:
        """Produce a dictionary with names, colspecs, and dtype for IGRA2 data.

        Returns a dict with entries 'body' and 'header'.
        """
        def _cdec(power=1):
            """Make a function to convert string 'value*10^power' to float."""
            def _cdec_power(val):
                if val in ['-9999', '-8888', '-99999']:
                    return np.nan
                else:
                    return float(val) / 10**power
            return _cdec_power

        def _cflag(val):
            """Replace alphabetic flags A and B with numeric."""
            if val == 'A':
                return 1
            elif val == 'B':
                return 2
            else:
                return 0

        def _ctime(strformat: str = 'MMMSS'):
            """Return a function converting a string from MMMSS or HHMM to seconds."""
            def _ctime_strformat(val):
                time = val.strip().zfill(5)

                if int(time) < 0:
                    return np.nan
                elif int(time) == 9999:
                    return np.nan
                else:
                    if strformat == 'MMMSS':
                        minutes = int(time[0:3])
                        seconds = int(time[3:5])
                        time_seconds = minutes * 60 + seconds
                    elif strformat == 'HHMM':
                        hours = int(time[0:2])
                        minutes = int(time[2:4])
                        time_seconds = hours * 3600 + minutes * 60
                    else:
                        sys.exit('Unrecognized time format')

                return time_seconds
            return _ctime_strformat

        def _clatlon(x):
            n = len(x)
            deg = x[0:n - 4]
            dec = x[n - 4:]
            return float(deg + '.' + dec)

        if self.suffix == '-drvd.txt':
            names_body = ['pressure', 'reported_height', 'calculated_height',
                          'temperature', 'temperature_gradient', 'potential_temperature',
                          'potential_temperature_gradient', 'virtual_temperature',
                          'virtual_potential_temperature', 'vapor_pressure',
                          'saturation_vapor_pressure', 'reported_relative_humidity',
                          'calculated_relative_humidity', 'relative_humidity_gradient',
                          'u_wind', 'u_wind_gradient', 'v_wind', 'v_wind_gradient',
                          'refractive_index']

            colspecs_body = [(0, 7), (8, 15), (16, 23), (24, 31), (32, 39),
                             (40, 47), (48, 55), (56, 63), (64, 71), (72, 79),
                             (80, 87), (88, 95), (96, 103), (104, 111), (112, 119),
                             (120, 127), (128, 135), (137, 143), (144, 151)]

            conv_body = {'pressure': _cdec(power=2),
                         'reported_height': int,
                         'calculated_height': int,
                         'temperature': _cdec(),
                         'temperature_gradient': _cdec(),
                         'potential_temperature': _cdec(),
                         'potential_temperature_gradient': _cdec(),
                         'virtual_temperature': _cdec(),
                         'virtual_potential_temperature': _cdec(),
                         'vapor_pressure': _cdec(power=3),
                         'saturation_vapor_pressure': _cdec(power=3),
                         'reported_relative_humidity': _cdec(),
                         'calculated_relative_humidity': _cdec(),
                         'relative_humidity_gradient': _cdec(),
                         'u_wind': _cdec(),
                         'u_wind_gradient': _cdec(),
                         'v_wind': _cdec(),
                         'v_wind_gradient': _cdec(),
                         'refractive_index': int}

            names_header = ['site_id', 'year', 'month', 'day', 'hour', 'release_time',
                            'number_levels', 'precipitable_water', 'inv_pressure',
                            'inv_height', 'inv_strength', 'mixed_layer_pressure',
                            'mixed_layer_height', 'freezing_point_pressure',
                            'freezing_point_height', 'lcl_pressure', 'lcl_height',
                            'lfc_pressure', 'lfc_height', 'lnb_pressure', 'lnb_height',
                            'lifted_index', 'showalter_index', 'k_index', 'total_totals_index',
                            'cape', 'convective_inhibition']

            colspecs_header = [(1, 12), (13, 17), (18, 20), (21, 23), (24, 26),
                               (27, 31), (31, 36), (37, 43), (43, 48), (49, 55),
                               (55, 61), (61, 67), (67, 73), (73, 79), (79, 85),
                               (85, 91), (91, 97), (97, 103), (103, 109), (109, 115),
                               (115, 121), (121, 127), (127, 133), (133, 139),
                               (139, 145), (145, 151), (151, 157)]

            conv_header = {'site_id': str,
                           'year': int,
                           'month': int,
                           'day': int,
                           'hour': int,
                           'release_time': _ctime(strformat='HHMM'),
                           'number_levels': int,
                           'precipitable_water': _cdec(power=2),
                           'inv_pressure': _cdec(power=2),
                           'inv_height': int,
                           'inv_strength': _cdec(),
                           'mixed_layer_pressure': _cdec(power=2),
                           'mixed_layer_height': int,
                           'freezing_point_pressure': _cdec(power=2),
                           'freezing_point_height': int,
                           'lcl_pressure': _cdec(power=2),
                           'lcl_height': int,
                           'lfc_pressure': _cdec(power=2),
                           'lfc_height': int,
                           'lnb_pressure': _cdec(power=2),
                           'lnb_height': int,
                           'lifted_index': int,
                           'showalter_index': int,
                           'k_index': int,
                           'total_totals_index': int,
                           'cape': int,
                           'convective_inhibition': int}

            na_vals = ['-99999']

        else:
            names_body = ['lvltyp1', 'lvltyp2', 'etime', 'pressure',
                          'pflag', 'height', 'zflag', 'temperature', 'tflag',
                          'relative_humidity', 'dewpoint_depression',
                          'direction', 'speed']

            colspecs_body = [(0, 1), (1, 2), (3, 8), (9, 15), (15, 16),
                             (16, 21), (21, 22), (22, 27), (27, 28),
                             (28, 33), (34, 39), (40, 45), (46, 51)]

            conv_body = {'lvltyp1': int,
                         'lvltyp2': int,
                         'etime': _ctime(strformat='MMMSS'),
                         'pressure': _cdec(power=2),
                         'pflag': _cflag,
                         'height': int,
                         'zflag': _cflag,
                         'temperature': _cdec(),
                         'tflag': _cflag,
                         'relative_humidity': _cdec(),
                         'dewpoint_depression': _cdec(),
                         'direction': int,
                         'speed': _cdec()}

            names_header = ['site_id', 'year', 'month', 'day', 'hour', 'release_time',
                            'number_levels', 'pressure_source_code',
                            'non_pressure_source_code',
                            'latitude', 'longitude']

            colspecs_header = [(1, 12), (13, 17), (18, 20), (21, 23), (24, 26),
                               (27, 31), (32, 36), (37, 45), (46, 54), (55, 62), (63, 71)]

            na_vals = ['-8888', '-9999']

            conv_header = {'release_time': _ctime(strformat='HHMM'),
                           'number_levels': int,
                           'latitude': _clatlon,
                           'longitude': _clatlon}

        return {'body': {'names': names_body,
                         'colspecs': colspecs_body,
                         'converters': conv_body,
                         'na_values': na_vals,
                         'index_col': False},
                'header': {'names': names_header,
                           'colspecs': colspecs_header,
                           'converters': conv_header,
                           'na_values': na_vals,
                           'index_col': False}}

    def _clean_body_df(self, df: pd.DataFrame):
        """Format the dataframe, remove empty rows, and add units attribute."""
        if self.suffix == '-drvd.txt':
            df = df.dropna(subset=('temperature', 'reported_relative_humidity',
                           'u_wind', 'v_wind'), how='all').reset_index(drop=True)

            df.attrs['units'] = {'pressure': 'hPa',
                        'reported_height': 'meter',
                        'calculated_height': 'meter',
                        'temperature': 'Kelvin',
                        'temperature_gradient': 'Kelvin / kilometer',
                        'potential_temperature': 'Kelvin',
                        'potential_temperature_gradient': 'Kelvin / kilometer',
                        'virtual_temperature': 'Kelvin',
                        'virtual_potential_temperature': 'Kelvin',
                        'vapor_pressure': 'Pascal',
                        'saturation_vapor_pressure': 'Pascal',
                        'reported_relative_humidity': 'percent',
                        'calculated_relative_humidity': 'percent',
                        'relative_humidity_gradient': 'percent / kilometer',
                        'u_wind': 'meter / second',
                        'u_wind_gradient': '(meter / second) / kilometer)',
                        'v_wind': 'meter / second',
                        'v_wind_gradient': '(meter / second) / kilometer)',
                        'refractive_index': 'unitless'}

        else:
            df['u_wind'], df['v_wind'] = get_wind_components(df['speed'],
                                                             np.deg2rad(df['direction']))
            df['u_wind'] = np.round(df['u_wind'], 1)
            df['v_wind'] = np.round(df['v_wind'], 1)

            df = df.dropna(subset=('temperature', 'direction', 'speed',
                           'dewpoint_depression', 'u_wind', 'v_wind'),
                           how='all').reset_index(drop=True)

            df['dewpoint'] = df['temperature'] - df['dewpoint_depression']

            df.drop('dewpoint_depression', axis=1, inplace=True)

            df.attrs['units'] = {'etime': 'second',
                        'pressure': 'hPa',
                        'height': 'meter',
                        'temperature': 'degC',
                        'dewpoint': 'degC',
                        'direction': 'degrees',
                        'speed': 'meter / second',
                        'u_wind': 'meter / second',
                        'v_wind': 'meter / second'}

        return df

    def _clean_header_df(self, df: pd.DataFrame) -> pd.DataFrame:
        """Format the header dataframe and add units."""
        if self.suffix == '-drvd.txt':
            df.attrs['units'] = {'release_time': 'second',
                        'precipitable_water': 'millimeter',
                        'inv_pressure': 'hPa',
                        'inv_height': 'meter',
                        'inv_strength': 'Kelvin',
                        'mixed_layer_pressure': 'hPa',
                        'mixed_layer_height': 'meter',
                        'freezing_point_pressure': 'hPa',
                        'freezing_point_height': 'meter',
                        'lcl_pressure': 'hPa',
                        'lcl_height': 'meter',
                        'lfc_pressure': 'hPa',
                        'lfc_height': 'meter',
                        'lnb_pressure': 'hPa',
                        'lnb_height': 'meter',
                        'lifted_index': 'degC',
                        'showalter_index': 'degC',
                        'k_index': 'degC',
                        'total_totals_index': 'degC',
                        'cape': 'Joule / kilogram',
                        'convective_inhibition': 'Joule / kilogram'}

        else:
            df.attrs['units'] = {'release_time': 'second',
                        'latitude': 'degrees',
                        'longitude': 'degrees'}

        return df


import os
from unittest import TestCase
import pytest
import numpy as np
import pandas as pd
from numpy.testing import assert_allclose, assert_equal
from pyrtlib.absorption_model import H2OAbsModel, O2AbsModel, O3AbsModel
from pyrtlib.uncertainty import AbsModUncertainty, SpectroscopicParameter
from pyrtlib.uncertainty import covariance_matrix
from pyrtlib.climatology import AtmosphericProfiles as atmp
from pyrtlib.tb_spectrum import TbCloudRTE
from pyrtlib.utils import ppmv2gkg, mr2rh, ppmv_to_moleculesm3

THIS_DIR = os.path.dirname(os.path.abspath(__file__))

class Test(TestCase):
    def test_absmod_uncertainties_perturb_water_min(self):
        w_sp = SpectroscopicParameter.water_parameters('R17')
        w_sp['gamma_a'].uncer[0:2] = np.array([0.039, 0.01])
        SpectroscopicParameter.set_parameters(w_sp)
        
        amu = AbsModUncertainty.parameters_perturbation(['gamma_a'], 'min', 0)
        expected0 = 2.626
        assert_equal(expected0, amu['gamma_a'].value[0])
        amu = AbsModUncertainty.parameters_perturbation(['gamma_a'], 'min', 1)
        expected1 = 2.926
        assert_equal(expected1, amu['gamma_a'].value[1])

    def test_absmod_uncertainties_perturb_oxygen_max(self):
        o_sp = SpectroscopicParameter.oxygen_parameters('R18')
        o_sp['O2S'].value[0:2] = np.array([2.906e-15, 7.957e-16])
        o_sp['O2S'].uncer[0:2] = o_sp['O2S'].value[0:2] / 100
        SpectroscopicParameter.set_parameters(o_sp)

        amu = AbsModUncertainty.parameters_perturbation(['O2S'], 'max', 0)
        expected0 = 2.93506e-15
        assert_equal(expected0, amu['O2S'].value[0])

        amu = AbsModUncertainty.parameters_perturbation(['O2S'], 'max', 1)
        expected1 = 8.03657e-16
        assert_equal(expected1, amu['O2S'].value[1])

    def test_absmod_uncertainties_perturb_ozono_max(self):
        oz_sp = SpectroscopicParameter.ozono_parameters('R18')
        oz_sp['O3_S1'].value[0:2] = np.array([2.272e-13, 3.547e-13])
        oz_sp['O3_S1'].uncer[0:2] = oz_sp['O3_S1'].value[0:2] / 100
        SpectroscopicParameter.set_parameters(oz_sp)

        amu = AbsModUncertainty.parameters_perturbation(['O3_S1'], 'max', 0)
        expected0 = 2.29472e-13
        assert_equal(expected0, amu['O3_S1'].value[0])

        amu = AbsModUncertainty.parameters_perturbation(['O3_S1'], 'max', 1)
        expected1 = 3.58247e-13
        assert_equal(expected1, amu['O3_S1'].value[1])

    def test_pyrtlib_uncertainty_gamma_a(self):
        z, p, _, t, md = atmp.gl_atm(atmp.TROPICAL)

        gkg = ppmv2gkg(md[:, atmp.H2O], atmp.H2O)
        rh = mr2rh(p, t, gkg)[0] / 100

        frq = np.arange(20, 201, 1)

        water_sp = SpectroscopicParameter.water_parameters("R17")
        oxygen_sp = SpectroscopicParameter.oxygen_parameters("R18")

        parameters = {**water_sp, **oxygen_sp}
        SpectroscopicParameter.set_parameters(parameters)

        amu = AbsModUncertainty.parameters_perturbation(
            ['gamma_a'], 'min', index=1)

        rte = TbCloudRTE(z, p, t, rh, frq, amu=amu)
        rte.init_absmdl('R17')
        O2AbsModel.model = 'R18'
        O2AbsModel.set_ll()
        df = rte.execute()

        df_expected = pd.read_csv(
            os.path.join(THIS_DIR, "data", "tbtotal_uncertainty_gamma_a_min.csv"))
        assert_allclose(df.tbtotal.values, df_expected.gamma_a)

    def test_pyrtlib_uncertainty_wO3(self):
        z, p, _, t, md = atmp.gl_atm(atmp.US_STANDARD)

        o3n_ppmv = md[:, atmp.O3]
        o3n = np.zeros(z.shape)
        for k in range(0, len(z)):
            o3n[k] = ppmv_to_moleculesm3(o3n_ppmv[k], p[k] * 100.0, t[k])

        gkg = ppmv2gkg(md[:, atmp.H2O], atmp.H2O)
        rh = mr2rh(p, t, gkg)[0] / 100

        frq = np.arange(20, 201, 1)

        water_sp = SpectroscopicParameter.water_parameters("R21SD")
        oxygen_sp = SpectroscopicParameter.oxygen_parameters("R20")
        ozono_sp = SpectroscopicParameter.ozono_parameters("R18")

        parameters = {**water_sp, **oxygen_sp, **ozono_sp}
        SpectroscopicParameter.set_parameters(parameters)

        amu = AbsModUncertainty.parameters_perturbation(
            ['O3_X'], 'min', index=0)

        rte = TbCloudRTE(z, p, t, rh, frq, o3n=o3n)
        rte.set_amu(amu)
        rte.init_absmdl('R20')
        H2OAbsModel.model = 'R21SD'
        H2OAbsModel.set_ll()
        O3AbsModel.model = 'R18'
        O3AbsModel.set_ll()
        df = rte.execute()

        df_expected = pd.read_csv(
            os.path.join(THIS_DIR, "data", "tbtotal_uncertainty_gamma_a_min.csv"))
        assert_allclose(df.tbtotal.values, df_expected.o3_x)

    def test_set_amu(self):
        z, p, _, t, md = atmp.gl_atm(atmp.US_STANDARD)

        gkg = ppmv2gkg(md[:, atmp.H2O], atmp.H2O)
        rh = mr2rh(p, t, gkg)[0] / 100

        frq = np.arange(20, 201, 1)
        water_sp = SpectroscopicParameter.water_parameters("R17")
        oxygen_sp = SpectroscopicParameter.oxygen_parameters("R18")
        water_sp['con_Cf_factr'].value = 1
        water_sp['con_Cs_factr'].value = 1
        parameters = {**water_sp, **oxygen_sp}
        SpectroscopicParameter.set_parameters(parameters)

        rte = TbCloudRTE(z, p, t, rh, frq)
        rte.satellite = False
        rte.init_absmdl('R17')
        O2AbsModel.model = 'R18'
        O2AbsModel.set_ll()
        df = rte.execute()

        rte.set_amu(parameters)
        df_expected = rte.execute()

        assert_allclose(df.tbtotal.values, df_expected.tbtotal.values)

    def test_cov_p(self):
        n = 106
        r111 = covariance_matrix.R17_111
        r112 = covariance_matrix.R17_112

        assert_allclose(r111[0:n, 0:n], r112[0:n, 0:n])

    # @pytest.mark.skip(reason="skipping")
    def test_spectroscopic_params(self):
        oxygen_sp = SpectroscopicParameter.oxygen_parameters("R18")
        v = oxygen_sp['w2a'].value
        assert_equal(v, 1.2)

    def test_add_spectr_params(self):
        parameters = SpectroscopicParameter.water_parameters("R17")
        parameters['test'] = SpectroscopicParameter(
            2.3, 0.001, 'unitless', 'Tretyakov, JMS, 2016')
        parameters['test'].value
        assert_equal(parameters['test'].value, 2.3)

    # @pytest.mark.skip(reason="skipping")
    def test_edit_spectroscopic_params(self):
        parameters = SpectroscopicParameter.oxygen_parameters("R18")
        parameters['w2a'].value = 1.4
        assert_equal(parameters['w2a'].value, 1.4)

    # @pytest.mark.skip(reason="skipping")
    def test_set_parameters(self):
        parameters = SpectroscopicParameter.water_parameters("R17")
        parameters['gamma_a'].value[0] = 2.688
        parameters['gamma_a'].uncer[0] = 0.039
        # SpectroscopicParameter.set_parameters(parameters)
        assert_equal(parameters['gamma_a'].value[0], 2.688)

    def test_uncertainty_propagation(self):
        parameters = SpectroscopicParameter.water_parameters("R19")
        parameters['delta_a'].uncer = np.array([0.005])
        parameters['gamma_a'].uncer = np.array([0.022])
        u = AbsModUncertainty.uncertainty_propagation(
            "A/B", parameters['delta_a'].value[0],
            parameters['gamma_a'].value[0],
            parameters['delta_a'].uncer[0],
            parameters['gamma_a'].uncer[0])
        expected = (-0.0122267506483882929, 0.0018552168412088639)
        assert_equal(expected, u)
        u = AbsModUncertainty.uncertainty_propagation(
            "aA", parameters['delta_a'].value[0],
            parameters['gamma_a'].value[0],
            parameters['delta_a'].uncer[0],
            parameters['gamma_a'].uncer[0])
        expected = (-0.033, 0.005)
        assert_equal(expected, u)
        u = AbsModUncertainty.uncertainty_propagation(
            "aA+bB", parameters['delta_a'].value[0],
            parameters['gamma_a'].value[0],
            parameters['delta_a'].uncer[0],
            parameters['gamma_a'].uncer[0])
        expected = (2.666, 0.02256102834535695)
        assert_equal(expected, u)




"""
.. note:: *Phil* The parameters that contribute most uncertainty are related to pressure-broadening: widths, mixing, & their T dependence. The intensities and their temperature dependence are obtained (at least in my model) from HITRAN2012. B2=(Ef+Ei)/2kTo,
    where Ef, Ei=upper and lower energy levels of the line; k=Boltzmann const; To=296K. Although HITRAN has uncertainties for six parameters, the energy level is not among them.  
    I expect that it does not contribute significant uncertainty to the model.

.. note:: *From Verdes et al 2005* For the pressure shifts, an estimated uncertainty for all lines belonging 
    to the same molecular species is assumed. This is a poor approximation as it is well known that the pressure shift is usually very different (with 
    possible change of sign) from one given transition to the next. Thus, the investigated pressure shift uncertainties were 300 kHz/Torr for 
    H2O and 50 kHz/Torr for O2 lines
        
.. note:: *From Makarov et al., 2011* "The fidelity of the new model to the spectrometer data is generally better than 2% between 54 and 65 GHz."
    "It is therefore important to distinguish between the uncertainty of the calculated absorption, and uncertainties in the values of the model?s coefficients. The coefficients are adjusted to fit measurements, and underestimation of some coefficients can be counterbalanced by excess values in other coefficients when absorption is calculated."

"""

import os
from copy import deepcopy
from dataclasses import dataclass
from typing import Tuple, Optional, Dict
from dataclasses import dataclass, field

import numpy as np
# import scipy.interpolate as si
from pyrtlib.absorption_model import H2OAbsModel, O2AbsModel, O3AbsModel
from pyrtlib.utils import constants

# TVEC = np.loadtxt(
#     open(os.path.join(PATH, "tbd", "Tvec.csv"), "rb"), delimiter=",")
# PVEC = np.loadtxt(
#     open(os.path.join(PATH, "tbd", "Pvec.csv"), "rb"), delimiter=",")
# FVEC = np.loadtxt(
#     open(os.path.join(PATH, "tbd", "Fvec.csv"), "rb"), delimiter=",")
# PR_EXT = np.loadtxt(
#     open(os.path.join(PATH, "tbd", "PR_ext.csv"), "rb"), delimiter=",")


@dataclass
class SpectroscopicParameter:
    """Absorption model uncertainties for the spectroscopic parameters

    Example:
        >>> from pyrtlib.uncertainty import SpectroscopicParameter
        >>> parameters = SpectroscopicParameter.water_parameters('R18')
        >>> parameters['con_Cf'].value
        5.95e-10

        New value may be added to parameters using :py:func:`~pyrtlib.uncertainty.SpectroscopicParameter` class as following

        >>> parameters['con_Xs'] = SpectroscopicParameter(2.3, 0.001, 'unitless', 'Tretyakov, JMS, 2016')
        >>> parameters['con_Xs'].value
        2.3
        >>> parameters['con_Xs'].uncer
        'unitless'
        >>> parameters['con_Xs'].refer
        'Tretyakov, JMS, 2016'

        Also, existing parameters may be modified as following

        >>> parameters['w2a'].value = 1.333
        >>> parameters['w2a'].value
        1.333

    Note:
        If new value will be added or modified it is necessary to save the new values by calling
        the :py:func:`~pyrtlib.uncertainty.SpectroscopicParameter.set_parameters` function.

    """
    value: np.ndarray
    """The value associated to the parameter
    """
    uncer: np.ndarray
    """The uncertainty of the parameter
    """
    units: Optional[str] = ''
    """The units of the parameter
    """
    refer: Optional[str] = ''
    """The reference of the parameter
    """
    name: Optional[str] = ''
    """The name or description of the parameter
    """

    def _initialize() -> Dict:
        SpectroscopicParameter._parameters()

    @staticmethod
    def water_parameters(model: str) -> Dict:
        """This method is used for uncertainty analysis and returns the dictionary
        with the whole spectroscopic parameters for :math:`H_2O`.

        Args:
            model (str): The absorption model.

        Returns:
            Dict: The :math:`H_2O` spectroscopic parameters dictionary

        Example:
            >>> sp = SpectroscopicParameter.water_parameters("R19")
            >>> sp['con_Xs']
            SpectroscopicParameter(value=7.5, 
                                   uncer=0.0, 
                                   units='unitless', 
                                   refer='Tretyakov, JMS, 2016', 
                                   name='Self broadening temperature dependence exponents')
        """
        H2OAbsModel.model = model
        H2OAbsModel.set_ll()
        ll = H2OAbsModel.h2oll

        h2o_sp = {
            "con_Cf": SpectroscopicParameter(value=ll.cf, uncer=.0, units='1/(km*(mb^2*GHz^2))', name='Foreign induced broadening coefficient'),
            "con_Cs": SpectroscopicParameter(value=ll.cs, uncer=.0, units='1/(km*(mb^2*GHz^2))', name='Self induced broadening coefficient'),
            
            "con_Xf": SpectroscopicParameter(value=ll.xcf, uncer=.0, units='unitless', name='Foreign broadening temperature dependence exponents'),
            "con_Xs": SpectroscopicParameter(value=ll.xcs, uncer=.0, units='unitless', name='Self broadening temperature dependence exponents'),

            "FL": SpectroscopicParameter(value=ll.fl, uncer=np.zeros(len(ll.fl)), units='GHz', name='Central frequency line'),
            "S": SpectroscopicParameter(value=ll.s1, uncer=np.zeros(len(ll.fl)), units='Hz*cm^2', name='Line intensity (or strength)'),
            "B2": SpectroscopicParameter(value=ll.b2, uncer=.0, units='unitless', name='Temperature coefficient of intensity'),

            "gamma_a": SpectroscopicParameter(value=ll.w0*1e3, uncer=np.zeros(len(ll.fl)), units='MHz/mb', name='Air induced broadening coefficients'),
            "gamma_w": SpectroscopicParameter(value=ll.w0s*1e3, uncer=np.zeros(len(ll.fl)), units='MHz/mb', name='Water induced broadening coefficients'),
            "n_a": SpectroscopicParameter(value=ll.x, uncer=np.zeros(len(ll.fl)), units='unitless', name='Temperature-dependence exponent of air'),
            "n_w": SpectroscopicParameter(value=ll.xs, uncer=np.zeros(len(ll.fl)), units='unitless', name='Temperature-dependence exponent of water'),

            "wv_nS": SpectroscopicParameter(value=.0, uncer=.0, name='Intensity temperature-dependence exponent'),
        }
        if H2OAbsModel.model > 'R17':
            h2o_sp_ = {
                "con_Cf_factr": SpectroscopicParameter(value=1, uncer=.0, units='unitless', refer='Turner et al., TGRSS, 2009', name=''),
                "con_Cs_factr": SpectroscopicParameter(value=1, uncer=.0, units='unitless', refer='Turner et al., TGRSS, 2009', name=''),
                "n_da": SpectroscopicParameter(value=ll.xh, uncer=np.zeros(len(ll.fl)), name='T-exponent of air-shifting'),
                "n_dw": SpectroscopicParameter(value=ll.xhs, uncer=np.zeros(len(ll.fl)), name='T-exponent of water-shifting'),
                "delta_a": SpectroscopicParameter(value=ll.sh*1e3, uncer=np.zeros(len(ll.fl)), units='MHz/mb', name='Air induced shifting coefficient'),
                "delta_w": SpectroscopicParameter(value=ll.shs*1e3, uncer=np.zeros(len(ll.fl)), units='MHz/mb', name='Water induced shifting coefficient'),
            }
        else:
            h2o_sp_ = {
                "SR": SpectroscopicParameter(value=ll.sr, uncer=np.zeros(len(ll.fl)), units='unitless', name='Shift to width ratio'), 
                "con_Cf_factr": SpectroscopicParameter(value=1.11, uncer=np.sqrt(0.098 ** 2 + 0.03 ** 2), units='unitless', refer='Turner et al., TGRSS, 2009', name=''),
                "con_Cs_factr": SpectroscopicParameter(value=0.79, uncer=np.sqrt(0.17 ** 2 + 0.06 ** 2), units='unitless', refer='Turner et al., TGRSS, 2009', name=''),}

        h2o_sp = {**h2o_sp, **h2o_sp_}

        return h2o_sp

    @staticmethod
    def oxygen_parameters(model: str) -> Dict:
        """This method is used for uncertainty analysis and returns the dictionary
        with the whole spectroscopic parameters for :math:`O_2`.

        Args:
            model (str): The absorption model.

        Returns:
            Dict: The :math:`O_2` spectroscopic parameters dictionary
        """
        O2AbsModel.model = model
        O2AbsModel.set_ll()
        ll = O2AbsModel.o2ll

        o2_sp = {
            "O2FL": SpectroscopicParameter(value=ll.f, uncer=.0, units='GHz', name='Line frequency'),
            "O2S": SpectroscopicParameter(value=ll.s300, uncer=np.zeros(len(ll.f)), units='cm2*Hz', name='Line intensity (or strength)'),
            "O2BE": SpectroscopicParameter(value=ll.be, uncer=.0, units='unitless', name='Temperature exponent for intensity'),
            "O2gamma": SpectroscopicParameter(value=ll.w300, uncer=np.zeros(len(ll.f)), units='MHz/mb', name='Self broadening temperature dependence exponents'),
            "O2gamma_NL": SpectroscopicParameter(value=ll.w300, uncer=.0, units='MHz/mb', name='Self broadening temperature dependence exponents for neglected lines (NL)'),
            "Snr": SpectroscopicParameter(value=1.6e-17, uncer=.0, units='Hz*cm2/GHz2', name='Line intensity of non-resonant pseudo-line'),
            "WB300": SpectroscopicParameter(value=ll.wb300, uncer=.0, units='MHz/mb', name='Pressure broadening of non-resonant pseudo-line'),
            "X11": SpectroscopicParameter(value=0.785, uncer=.0, units='unitless', name='Temperature dependence of broadening coefficient'),
            "X16": SpectroscopicParameter(value=0.765, uncer=.0, units='unitless', name='Temperature dependence of broadening coefficient'),
            "X05": SpectroscopicParameter(value=ll.x, uncer=.0, units='unitless', name='Temperature dependence of broadening coefficient'),
            "O2_nS": SpectroscopicParameter(value=2.0, uncer=.0, units='unitless',  name='Intensity temperature-dependence exponent'),
            "w2a": SpectroscopicParameter(value=1.2, uncer=.0, units='unitless', name='Water-to-air broadening ratio'),
        }
        if O2AbsModel.model > 'R19':
            o2_sp_ = {
                "y0": SpectroscopicParameter(value=ll.y0, uncer=.0, name=''),
                "y1": SpectroscopicParameter(value=ll.y1, uncer=.0, name=''),
                "dnu0": SpectroscopicParameter(value=ll.dnu0, uncer=.0, name=''),
                "dnu1": SpectroscopicParameter(value=ll.dnu1, uncer=.0, name=''),
            }
        else:
            o2_sp_ = {
                "O2_V": SpectroscopicParameter(value=ll.v, uncer=np.zeros(len(ll.f)), units='1/bar == 1/1e5Pa =~ 1/atm', name='O2 line mixing temperature dependence'),
                "O2_V_NL": SpectroscopicParameter(value=ll.v, uncer=.0, units='1/mb', name='O2 line mixing temperature dependence for neglected lines (NL)'),
                "Y300": SpectroscopicParameter(value=ll.y300, uncer=np.zeros(len(ll.f)), units='1/bar == 1/1e5Pa =~ 1/atm', name='Line mixing coefficients (single lines)'),
                "Y300_NL": SpectroscopicParameter(value=ll.y300, uncer=.0, units='1/mb', name='Line mixing coefficients for neglected lines (NL)'),
            }
        o2_sp = {**o2_sp, **o2_sp_}

        return o2_sp

    @staticmethod
    def ozono_parameters(model: str) -> Dict:
        """This method is used for uncertainty analysis and returns the dictionary
        with the whole spectroscopic parameters for :math:`O_3`.

        Args:
            model (str): The absorption model.

        Returns:
            Dict: The Ozono (:math:`O_3`) spectroscopic parameters dictionary
        """
        O3AbsModel.model = model
        O3AbsModel.set_ll()
        ll = O3AbsModel.o3ll

        o3_sp = {
            "O3_FL": SpectroscopicParameter(value=ll.fl, uncer=np.zeros(len(ll.fl)), units='GHz', name='Line frequency'),
            "O3_S1": SpectroscopicParameter(value=ll.s1, uncer=np.zeros(len(ll.fl)), units='Hz*cm^2', name='Line intensity (or strength)'),
            "O3_B": SpectroscopicParameter(value=ll.b, uncer=np.zeros(len(ll.fl)), units='unitless', name='Temperature Coefficient of intensity'),
            "O3_W": SpectroscopicParameter(value=ll.w, uncer=np.zeros(len(ll.fl)), units='GHz/mb', name='Air-pressure broadening'),
            "O3_X": SpectroscopicParameter(value=ll.x, uncer=np.zeros(len(ll.fl)), units='unitless', name='Temperature exponent of air broadening'),
            "O3_SR": SpectroscopicParameter(value=ll.sr, uncer=np.zeros(len(ll.fl)), units='unitless', name='Shift-to-width ratio'),
        }

        return o3_sp

    @staticmethod
    def set_parameters(SP: dict) -> None:
        """Set new values and uncertainties to spectroscopic parameters.

        Args:
            SP (dict): The new dictionary with the spectroscopic parameters

        Example:
            >>> parameters = SpectroscopicParameter.water_parameters('R17')
            >>> parameters['gamma_a'].value[0] = 2.688
            >>> parameters['gamma_a'].uncer[0] = 0.039
            >>> SpectroscopicParameter.set_parameters(parameters)

        """
        SpectroscopicParameter.SP = SP


class AbsModUncertainty:
    """This module provides the uncertainties affecting absorption model
    coefficients found in the litterature.
    The baseline are the routines of Rosenkranz 2016 + modification to water-2-air by [Koshelev-2015]_.

    For a more exaustive example on how uncertainty work look at :ref:`uncert_example`.

    References:
        .. [1] [Koshelev-2011]_
        .. [2] [Koshelev-2015]_
        .. [3] [Koshelev-2018]_
        .. [4] [Koshelev-2017]_
        .. [5] [Turner-2009]_
        .. [6] [Tretyakov-2016]_
    """

    @staticmethod
    def uncertainty_propagation(fun: str, A: np.ndarray, B: np.ndarray, sA: np.ndarray, sB: np.ndarray,
                                a: Optional[float] = 1.0, b: Optional[float] = 1.0, sAB: Optional[float] = 0.0) -> \
            Tuple[np.ndarray, np.ndarray]:
        r"""This function propagates uncertainty given two variables A, B and their
        associated uncertainty.

        +--------------------------------------------+------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        | Function                                   |  Standard deviation                                                                                                                                                                |
        +--------------------------------------------+------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        | :math:`f=aA`                               |  :math:`\sigma_{f}=|a|\sigma _{A}`                                                                                                                                                 |
        +--------------------------------------------+------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        | :math:`\displaystyle f=aA+bB`              | :math:`\displaystyle \sigma _{f}={\sqrt {a^{2}\sigma _{A}^{2}+b^{2}\sigma _{B}^{2}+2ab\,\sigma _{AB}}}`                                                                            |
        +--------------------------------------------+------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        | :math:`\displaystyle f=aA-bB`              | :math:`\displaystyle \sigma _{f}={\sqrt {a^{2}\sigma _{A}^{2}+b^{2}\sigma _{B}^{2}-2ab\,\sigma _{AB}}}`                                                                            |
        +--------------------------------------------+------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        | :math:`\displaystyle f=AB`                 | :math:`\displaystyle \sigma _{f}\approx \left|f\right|{\sqrt {\left({\frac {\sigma _{A}}{A}}\right)^{2}+\left({\frac {\sigma _{B}}{B}}\right)^{2}+2{\frac {\sigma _{AB}}{AB}}}}`   |
        +--------------------------------------------+------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        | :math:`\displaystyle f={\frac {A}{B}}`     | :math:`\displaystyle \sigma _{f}\approx \left|f\right|{\sqrt {\left({\frac {\sigma _{A}}{A}}\right)^{2}+\left({\frac {\sigma _{B}}{B}}\right)^{2}-2{\frac {\sigma _{AB}}{AB}}}}`   |
        +--------------------------------------------+------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        | :math:`\displaystyle f=A^{a}`              | :math:`\displaystyle \sigma _{f}=|a|A^{a-1}\sigma _{A}`                                                                                                                            |
        +--------------------------------------------+------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+

        Args:
            fun (str): [description].
            A (numpy.ndarray): variable A
            B (numpy.ndarray): variable B
            sA (numpy.ndarray): uncertainty on A
            sB (numpy.ndarray): uncertainty on B
            a (float): multiplier for A. Defaults to 1.0.
            b (float): multiplier for B. Defaults to 1.0.
            sAB (float): sqrt of covariance between A and B. Defaults to 0.0.

        Raises:
            ValueError: Raise if `fun` argument is not defined.

        Returns:
            tuple: function computed with imput variables and uncertainty on function used

        References:
            .. [1] Wikipedia: Propagation of uncertainty https://en.wikipedia.org/wiki/Propagation_of_uncertainty
        """
        if fun == 'aA':
            F = a * A
            sF = np.abs(a) * sA
        elif fun == 'aA+bB':
            F = a * A + b * B
            sF = np.sqrt(a ** 2 * sA ** 2 + b ** 2 * sB ** 2 + 2 * a * b * sAB)
        elif fun == 'aA-bB':
            F = a * A - b * B
            sF = np.sqrt(a ** 2 * sA ** 2 + b ** 2 * sB ** 2 - 2 * a * b * sAB)
        elif fun == 'A*B':
            F = A * B
            # sF = np.sqrt( B**2*sA**2 + A**2*sB**2 + 2*A*B*sAB )
            sF = np.abs(A * B) * np.sqrt((sA / A) ** 2 +
                                         (sB / B) ** 2 + 2 * sAB / (A * B))
        elif fun == 'A/B':
            F = A / B
            sF = np.abs(A / B) * np.sqrt((sA / A) ** 2 +
                                         (sB / B) ** 2 - 2 * sAB / (A * B))
        elif fun == 'A^a':
            F = A ** a
            sF = np.abs(a) * A ** (a - 1) * sA
        else:
            raise ValueError("[utils] fun argument is mandatory")

        return F, sF

    @staticmethod
    def parameters_perturbation(what: Optional[list] = [], mode: Optional[str] = 'non', index: Optional[int] = None) -> Dict:
        """This functoin execute the perturbatoin of the spectroscopic parameters provided.

        Args:
            what (Optional[list], optional): The name of the parameter/s being perturbed. Defaults to [].
            mode (Optional[str], optional): The type of perturbation (max or min). Defaults to 'non'.
            index (Optional[int], optional): The index of the coefficient being perturbed. Defaults to None.

        Raises:
            ValueError: If argument `mode` is not set or invalid

        Returns:
            dict: The new dictionary that includes perturbed parameters.
        """
        try:
            amu_copy = deepcopy(SpectroscopicParameter.SP)
        except AttributeError as e:
            print('Spectroscopic parameters are not defined')
            raise e
        if what == []:
            param = list(amu_copy.keys())
        else:
            param = what

        # remember here we are using matlab logig for indexing (i.e. indes start from 1 instead that 0)
        # must be changed!!!!!
        param_index = index  # - 1 if index else index

        npar = len(param)
        for i in range(npar):
            new_value = amu_copy[param[i]].value
            uncer = amu_copy[param[i]].uncer
            if mode == 'max':
                # new_value[param_index] = new_value[param_index] + uncer[param_index]
                if isinstance(new_value, float) or isinstance(uncer, float):
                    amu_copy[param[i]].value = new_value + uncer
                else:
                    amu_copy[param[i]].value[param_index] = new_value[param_index] + \
                        uncer[param_index]
            elif mode == 'min':
                # new_value[param_index] = new_value[param_index] - uncer[param_index]
                if isinstance(new_value, float) or isinstance(uncer, float):
                    amu_copy[param[i]].value = new_value - uncer
                else:
                    amu_copy[param[i]].value[param_index] = new_value[param_index] - \
                        uncer[param_index]
            elif mode == 'non':
                pass
            elif mode == 'ran':
                # here rand or randn shall be used?
                # note that rand is only positive ([0-1]) and randn gives more
                # weight to values close to zero.
                # So, one way could be to combine them, using rand and the sign of randn
                # Assuming uncorrelated uncertainty:
                # for ip1 = param_indx
                #     p1.value(ip1) = p1.value(ip1) + p1.uncer(ip1) * rand(1) * sign(randn(1));
                # end
                raise NotImplementedError(
                    "sorry, random is not implemented yet")
            else:
                raise ValueError("mode args is not set or invalid")

        return amu_copy

    # @NotImplemented
    # @staticmethod
    # def _apu_line_mixing(f: np.ndarray, t: np.ndarray, p: np.ndarray, residual_source: Optional[int] = 0):
    #     unkn = 0.0  # unknown uncertainty outside the temp and freq range
    #     if residual_source == 1:
    #         pass
    #     elif residual_source == 2:
    #         pass
    #     else:
    #         tgrid, fgrid, pgrid = np.meshgrid(TVEC, FVEC, PVEC)
    #         apu = si.interpn((tgrid, fgrid, pgrid), PR_EXT, t, f, p, 'linear')

    #     return apu



import os
# from pathlib import Path
from unittest import TestCase
import pytest
import numpy as np
import pandas as pd
from datetime import datetime
from numpy.testing import assert_allclose, assert_equal
from pyrtlib.apiwebservices import WyomingUpperAir, ERA5Reanalysis, IGRAUpperAir

# TEST_DIR = Path(__file__).parent
# DATA_DIR = os.path.join(TEST_DIR, 'data')
THIS_DIR = os.path.dirname(os.path.abspath(__file__))


class Test(TestCase):

    # @pytest.mark.datafiles(DATA_DIR)
    def test_wyom_get_stations(self):
        df = WyomingUpperAir.get_stations('pac')
        b = df[df.station_id=='45004'].station_name == 'Kings Park'

        assert b.values



"""Read upper air data from the Wyoming archives.
The orignal code of this class can be found in https://github.com/Unidata/siphon
Terms and conditions are as for siphon library license (https://github.com/Unidata/siphon/blob/master/LICENSE)
"""

import warnings
from datetime import datetime
from io import StringIO
from typing import Union, Optional

import pandas as pd
try:
    from bs4 import BeautifulSoup
except ModuleNotFoundError as e:
    warnings.warn("Module BS4 must be installed to download Wyoming RAOB dataset.")
from ..apiwebservices.webservices import HTTPEndPoint

warnings.filterwarnings('ignore', 'Pandas doesn\'t allow columns to be created', UserWarning)


class WyomingUpperAir(HTTPEndPoint):
    """Download and parse data from the University of Wyoming's upper air archive."""

    def __init__(self) -> None:
        """Set up endpoint."""
        super().__init__('http://weather.uwyo.edu/cgi-bin/sounding')

    @classmethod
    def request_data(cls, time: datetime, site_id: Union[str, int], **kwargs) -> pd.DataFrame:
        """Retrieve upper air observations from the Wyoming archive.

        Args:
            time (datetime.datetime): The date and time of the desired observation.
            site_id (Union[str, int]): The three letter ICAO identifier of the station for which data should be
                downloaded.

        Returns:
            pandas.DataFrame:  A dataframe containing the data
        """

        endpoint = cls()
        df = endpoint._get_data(time, site_id)
        return df

    @classmethod
    def get_stations(cls, region: Optional[str] = 'europe') -> pd.DataFrame:
        """Retrieve list of available stations from the Wyoming archive.

        +---------+----------------+
        | region  | name           |
        +---------+----------------+
        | naconf  | North America  |
        +---------+----------------+
        | samer   | South America  |
        +---------+----------------+
        | pac     | South Pacific  |
        +---------+----------------+
        | nz      | New Zealand    |
        +---------+----------------+
        | ant     | Antarctica     |
        +---------+----------------+
        | np      | Arctic         |
        +---------+----------------+
        | europe  | Europe         |
        +---------+----------------+
        | africa  | Africa         |
        +---------+----------------+
        | seasia  | Southeast Asia |
        +---------+----------------+
        | mideast | Mideast        |
        +---------+----------------+

        Args:
            region (Optional[str], optional): The name of region from which to get stations list. Defaults to 'europe'.

        Returns:
            pandas.DataFrame: A dDataFrame of stations id and name
        """

        endpoint = cls()
        endpoint._base = 'http://weather.uwyo.edu/upperair/'

        resp = endpoint._get_path('{}.html'.format(region))
        soup = BeautifulSoup(resp.text, 'html.parser')
        area_tag = soup.find_all('area')
        stations = {}
        for t in area_tag:
            title = t.get('title')
            if title:
                cod, name = title.split(' ')[0], ' '.join(title.split(' ')[1:]).strip()
                stations[cod] = name

        df = pd.DataFrame(list(stations.items()), columns=['station_id', 'station_name'])
        return df

    def _get_data(self, time: datetime, site_id: Union[str, int]) -> pd.DataFrame:
        r"""Download and parse upper air observations from an online archive.

        Parameters
        ----------
        time : datetime
            The date and time of the desired observation.

        site_id : str
            The three letter ICAO identifier of the station for which data should be
            downloaded.

        Returns
        -------
            df :  pandas.DataFrame
                containing the data

        """
        raw_data = self._get_data_raw(time, site_id)
        soup = BeautifulSoup(raw_data, 'html.parser')
        tabular_data = StringIO(soup.find_all('pre')[0].contents[0])
        title = soup.find_all('h2')[0].contents[0]
        col_names = ['pressure', 'height', 'temperature', 'dewpoint', 'rh', 'mixr']
        # df = pd.read_fwf(tabular_data, skiprows=5, usecols=[0, 1, 2, 3, 4, 5], names=col_names, infer_rows=300)
        # change to read_csv function as some data column (e.g. height) couls be corrupted
        # To be further investigated in order to catch the issue
        df = pd.read_csv(tabular_data, header=None, skiprows=5, delim_whitespace=True, usecols=[0, 1, 2, 3, 4, 5], names=col_names)
        # Drop any rows with all NaN values for T, Td, winds
        df = df.dropna(subset=('temperature', 'dewpoint'), how='any').reset_index(drop=True)

        # Parse metadata
        meta_data = soup.find_all('pre')[1].contents[0]
        lines = meta_data.splitlines()

        # If the station doesn't have a name identified we need to insert a
        # record showing this for parsing to proceed.
        if 'Station number' in lines[1]:
            lines.insert(1, 'Station identifier: ')

        station = lines[1].split(':')[1].strip()
        station_number = int(lines[2].split(':')[1].strip())
        sounding_time = datetime.strptime(lines[3].split(':')[1].strip(), '%y%m%d/%H%M')
        latitude = float(lines[4].split(':')[1].strip())
        longitude = float(lines[5].split(':')[1].strip())
        elevation = float(lines[6].split(':')[1].strip())

        df['station'] = station
        df['station_number'] = station_number
        df['time'] = sounding_time
        df['latitude'] = latitude
        df['longitude'] = longitude
        df['elevation'] = elevation
        df['title'] = title

        # Add unit dictionary
        df.units = {'pressure': 'hPa',
                    'height': 'meter',
                    'temperature': 'degC',
                    'dewpoint': 'degC',
                    'rh': '%',
                    'mixr': 'g/kg',
                    'station': None,
                    'station_number': None,
                    'time': None,
                    'latitude': 'degrees',
                    'longitude': 'degrees',
                    'elevation': 'meter'}
        return df

    def _get_data_raw(self, time: datetime, site_id: Union[str, int]) -> str:
        """Download data from the University of Wyoming's upper air archive.

        Parameters
        ----------
        time : datetime
            Date and time for which data should be downloaded
        site_id : str
            Site id for which data should be downloaded

        Returns
        -------
        text of the server response

        """
        path = ('?region=naconf&TYPE=TEXT%3ALIST'
                '&YEAR={time:%Y}&MONTH={time:%m}&FROM={time:%d%H}&TO={time:%d%H}'
                '&STNM={stid}').format(time=time, stid=site_id)

        resp = self._get_path(path)
        # See if the return is valid, but has no data
        if resp.text.find('Can\'t') != -1:
            raise ValueError(
                'No data available for {time:%Y-%m-%d %HZ} '
                'for station {stid}.'.format(time=time, stid=site_id))

        return resp.text



"""
Read and Download data from ERA5Reanalysis Reanalysis model data.
To download ERA5 dataset it is necessary to configure an API key.
Step-by-step to create an API key can be found to https://cds.climate.copernicus.eu/api-how-to
"""

__author__ = ''
__date__ = 'March 2021'
__copyright__ = '(C) 2021, CNR-IMAA'

import os
import math
import warnings
from datetime import datetime
from typing import Optional, Tuple

try:
    import cdsapi
except ModuleNotFoundError as e:
    warnings.warn(
        "Module CDSAPI must be installed to download ERA5 Reanalysis dataset.")
import numpy as np
# from scipy.spatial import cKDTree
from sklearn.neighbors import BallTree, KDTree
import pandas as pd
from netCDF4 import Dataset

from ..utils import atmospheric_tickness


class ERA5Reanalysis:
    """Read and Download data from ERA5 CDS Reanalysis model data"""

    @classmethod
    def read_data(cls, file: str, lonlat: tuple) -> pd.DataFrame:
        """Read data from the ERA5 Reanalysis dataset.

        Args:
            file (str): The netcdf file
            lonlat (tuple): longitude and latitude

        Returns:
            pandas.DataFrame: [description]

        .. note:: To convert specific cloud water content (CLWC) or specific cloud ice water content (CIWC) 
            from kg kg-1 to g m-3 using this function :py:meth:`pyrtlib.utils.kgkg_to_gm3`
        """
        ERAFIVE = cls()
        nc = Dataset(file)
        lats = nc.variables['latitude'][:]
        lons = nc.variables['longitude'][:]
        # idx_lat = ERAFIVE.find_nearest(lats, lonlat[1])
        # idx_lon = ERAFIVE.find_nearest(lons, lonlat[0])
        idx, dist = ERAFIVE._find_nearest(lons, lats, lonlat)

        pres = np.asarray(nc.variables['level'][:])
        temp = np.asarray(nc.variables['t'][:, :, idx, idx])
        # RH in decimal
        rh = np.asarray(nc.variables['r'][:, :, idx, idx]) / 100
        clwc = np.asarray(nc.variables['clwc'][:, :, idx, idx])
        ciwc = np.asarray(nc.variables['ciwc'][:, :, idx, idx])
        crwc = np.asarray(nc.variables['crwc'][:, :, idx, idx])
        cswc = np.asarray(nc.variables['cswc'][:, :, idx, idx])
        ozone = np.asarray(nc.variables['o3'][:, :, idx, idx])
        q = np.asarray(nc.variables['q'][:, :, idx, idx])

        z = atmospheric_tickness(np.flip(pres), np.flip(temp[0]), np.flip(q[0]))  # Altitude in km

        date = pd.to_datetime(
            nc.variables['time'][:], origin='1900-01-01 00:00:00.0', unit='h')

        df = pd.DataFrame({'p': np.flip(pres),
                           'z': z,
                           't': np.flip(temp[0]),
                           'rh': np.flip(rh[0]),
                           'clwc': np.flip(clwc[0]),
                           'ciwc': np.flip(ciwc[0]),
                           'crwc': np.flip(crwc[0]),
                           'cswc': np.flip(cswc[0]),
                           'o3': np.flip(ozone[0]),
                           'q': np.flip(q[0]),
                           'time': np.repeat(date, len(pres))
                           })
        
        return df  # , (idx, dist), np.stack((lons, lats))

    def _find_nearest(self, lons, lats, point):
        """Find index of nearest coordinate

        Args:
            array ([type]): [description]
            value ([type]): [description]

        Returns:
            [type]: [description]
        """
        # idx = np.searchsorted(array, value, side="left")
        # if np.amax(array) < value < np.amin(array):
        #     warnings.warn('Out of boundingbox: value {} is out of dataset extent'.format(value))
        # if idx > 0 and (idx == len(array) or math.fabs(value - array[idx - 1]) < math.fabs(value - array[idx])):
        #     return idx - 1
        # else:
        #     return idx

        # scipy
        # combined_x_y_arrays = np.dstack([lats.ravel(), lons.ravel()])[0]
        # # points = list(point.transpose())
        #
        # mytree = cKDTree(combined_x_y_arrays)
        # dist, indexes = mytree.query(point)

        ball = BallTree(np.radians(
            np.stack((lats, lons), axis=-1)), metric='haversine')
        dist, indexes = ball.query(np.radians(
            np.array(np.flip(point)).reshape(1, 2)), k=1)

        return indexes[0][0], dist

    @staticmethod
    def request_data(path: str, time: datetime, lonlat: tuple, resolution: Optional[float] = 0.25, offset: Optional[float] = 0.4) -> str:
        """Download ERA5Reanalysis data from the Copernicus Climate Change Service.

        Args:
            path (str): The output directory
            time (datetime): The date and time of the desired observation.
            lonlat (tuple): The coordinatre in degrees, longitude and latitude
            resolution (Optional[float], optional): The pixel size of the requested grid data. Defaults to 0.25.
            offset (Optional[float], optional): The offset to apply to coordinates to get the extent. Defaults to 0.3.

        Returns:
            str: The path to downloaded netcdf file
        """
        # North, West, South, Est
        extent = [lonlat[1] + offset, lonlat[0] - offset,
                  lonlat[1] - offset, lonlat[0] + offset]
        nc_file_name = 'era5_reanalysis-{}.nc'.format(time.isoformat())
        nc_file = os.path.join(path, nc_file_name)

        variables = ['relative_humidity', 'specific_cloud_ice_water_content', 'specific_cloud_liquid_water_content',
                     'specific_humidity', 'specific_rain_water_content', 'specific_snow_water_content', 'ozone_mass_mixing_ratio', 'temperature']
        c = cdsapi.Client()
        c.retrieve(
            'reanalysis-era5-pressure-levels',
            {
                'product_type': 'reanalysis',
                'variable': variables,
                'pressure_level': [
                    '1', '2', '3', '5', '7', '10', '20', '30', '50', '70', '100', '125', '150',
                    '175', '200', '225', '250', '300', '350', '400', '450', '500', '550', '600',
                    '650', '700', '750', '775', '800', '825', '850', '875', '900', '925', '950',
                    '975', '1000',
                ],
                'year': time.year,
                'month': time.month,
                'day': time.day,
                'time': '{}:00'.format(time.hour),
                'area': extent,
                'grid': [resolution, resolution],
                'format': 'netcdf',
            },
            nc_file)

        return nc_file



import os
# from pathlib import Path
from unittest import TestCase
import pytest
import numpy as np
import pandas as pd
from datetime import datetime
from numpy.testing import assert_allclose
from pyrtlib.absorption_model import H2OAbsModel, LiqAbsModel, O2AbsModel, O3AbsModel
from pyrtlib.climatology import AtmosphericProfiles as atmp
from pyrtlib.tb_spectrum import TbCloudRTE
from pyrtlib.apiwebservices import WyomingUpperAir, ERA5Reanalysis, IGRAUpperAir
from pyrtlib.utils import ppmv2gkg, ppmv_to_moleculesm3, mr2rh, import_lineshape, dewpoint2rh, kgkg_to_kgm3, pressure_to_height, to_kelvin

# TEST_DIR = Path(__file__).parent
# DATA_DIR = os.path.join(TEST_DIR, 'data')
THIS_DIR = os.path.dirname(os.path.abspath(__file__))


class Test(TestCase):

    # @pytest.mark.datafiles(DATA_DIR)
    def test_pyrtlib_sat_R19SD_atm(self):
        d = atmp.atm_profiles()

        for k, v in d.items():
            z, p, _, t, md = atmp.gl_atm(k)

            gkg = ppmv2gkg(md[:, atmp.H2O], atmp.H2O)
            rh = mr2rh(p, t, gkg)[0] / 100

            ang = np.array([90.])
            frq = np.arange(20, 201, 1)

            rte = TbCloudRTE(z, p, t, rh, frq, ang)
            rte.init_absmdl('R19SD')
            df = rte.execute()

            df_expected = pd.read_csv(os.path.join(THIS_DIR, "data", "tbtotal_atm_rose19sd.csv"))
            assert_allclose(df.tbtotal, df_expected[v.lower().replace(" ", "_")], atol=0)

    # @pytest.mark.datafiles(DATA_DIR)
    @pytest.mark.skip(reason="skipping")
    def test_pyrtlib_sat_R19SD(self):
        z, p, _, t, md = atmp.gl_atm(atmp.TROPICAL)

        gkg = ppmv2gkg(md[:, atmp.H2O], atmp.H2O)
        rh = mr2rh(p, t, gkg)[0] / 100

        ang = np.array([90.])
        frq = np.arange(20, 201, 1)

        rte = TbCloudRTE(z, p, t, rh, frq, ang)
        rte.init_absmdl('R19SD')
        df = rte.execute()

        df_expected = pd.read_csv(
            os.path.join(THIS_DIR, "data", "tb_tot_ros03_16_17_18_19_19sd_20_20sd_98_mak11_21sd.csv"))
        assert_allclose(df.tbtotal, df_expected.ros19sd, atol=0)

    # @pytest.mark.datafiles(DATA_DIR)
    @pytest.mark.skip(reason="skipping")
    def test_pyrtlib_sat_R19(self):
        z, p, _, t, md = atmp.gl_atm(atmp.TROPICAL)

        gkg = ppmv2gkg(md[:, atmp.H2O], atmp.H2O)
        rh = mr2rh(p, t, gkg)[0] / 100

        ang = np.array([90.])
        frq = np.arange(20, 201, 1)

        rte = TbCloudRTE(z, p, t, rh, frq, ang)
        rte.init_absmdl('R19')
        df = rte.execute()

        df_expected = pd.read_csv(
            os.path.join(THIS_DIR, "data", "tb_tot_ros03_16_17_18_19_19sd_20_20sd_98_mak11_21sd.csv"))
        assert_allclose(df.tbtotal, df_expected.ros19, atol=0)

    # @pytest.mark.datafiles(DATA_DIR)
    @pytest.mark.skip(reason="skipping")
    def test_pyrtlib_sat_R16(self):
        z, p, _, t, md = atmp.gl_atm(atmp.TROPICAL)

        gkg = ppmv2gkg(md[:, atmp.H2O], atmp.H2O)
        rh = mr2rh(p, t, gkg)[0] / 100

        ang = np.array([90.])
        frq = np.arange(20, 201, 1)

        rte = TbCloudRTE(z, p, t, rh, frq, ang)
        rte.init_absmdl('R16')
        df = rte.execute()

        df_expected = pd.read_csv(
            os.path.join(THIS_DIR, "data", "tb_tot_ros03_16_17_18_19_19sd_20_20sd_98_mak11_21sd.csv"))
        assert_allclose(df.tbtotal, df_expected.ros16, atol=0)

    # @pytest.mark.datafiles(DATA_DIR)
    @pytest.mark.skip(reason="skipping")
    def test_pyrtlib_sat_makarov11(self):
        z, p, _, t, md = atmp.gl_atm(atmp.TROPICAL)

        gkg = ppmv2gkg(md[:, atmp.H2O], atmp.H2O)
        rh = mr2rh(p, t, gkg)[0] / 100

        ang = np.array([90.])
        frq = np.arange(20, 201, 1)

        rte = TbCloudRTE(z, p, t, rh, frq, ang)
        rte.init_absmdl('R16')
        O2AbsModel.model = 'makarov11'
        O2AbsModel.set_ll()
        df = rte.execute()

        df_expected = pd.read_csv(
            os.path.join(THIS_DIR, "data", "tb_tot_ros03_16_17_18_19_19sd_20_20sd_98_mak11_21sd.csv"))
        assert_allclose(df.tbtotal, df_expected.mak11, atol=0)

    # @pytest.mark.datafiles(DATA_DIR)
    @pytest.mark.skip(reason="skipping")
    def test_pyrtlib_sat_R03(self):
        z, p, _, t, md = atmp.gl_atm(atmp.TROPICAL)

        gkg = ppmv2gkg(md[:, atmp.H2O], atmp.H2O)
        rh = mr2rh(p, t, gkg)[0] / 100

        ang = np.array([90.])
        frq = np.arange(20, 201, 1)

        rte = TbCloudRTE(z, p, t, rh, frq, ang)
        rte.init_absmdl('R03')
        df = rte.execute()

        df_expected = pd.read_csv(
            os.path.join(THIS_DIR, "data", "tb_tot_ros03_16_17_18_19_19sd_20_20sd_98_mak11_21sd.csv"))
        assert_allclose(df.tbtotal, df_expected.ros03, atol=0)

    # @pytest.mark.datafiles(DATA_DIR)
    @pytest.mark.skip(reason="skipping")
    def test_pyrtlib_sat_R17(self):
        z, p, _, t, md = atmp.gl_atm(atmp.TROPICAL)

        gkg = ppmv2gkg(md[:, atmp.H2O], atmp.H2O)
        rh = mr2rh(p, t, gkg)[0] / 100

        ang = np.array([90.])
        frq = np.arange(20, 201, 1)

        rte = TbCloudRTE(z, p, t, rh, frq, ang)
        rte.init_absmdl('R17')
        df = rte.execute()

        df_expected = pd.read_csv(
            os.path.join(THIS_DIR, "data", "tb_tot_ros03_16_17_18_19_19sd_20_20sd_98_mak11_21sd.csv"))
        assert_allclose(df.tbtotal, df_expected.ros17, atol=0)

    # @pytest.mark.datafiles(DATA_DIR)
    @pytest.mark.skip(reason="skipping")
    def test_pyrtlib_sat_R20(self):
        z, p, _, t, md = atmp.gl_atm(atmp.TROPICAL)

        gkg = ppmv2gkg(md[:, atmp.H2O], atmp.H2O)
        rh = mr2rh(p, t, gkg)[0] / 100

        ang = np.array([90.])
        frq = np.arange(20, 201, 1)

        rte = TbCloudRTE(z, p, t, rh, frq, ang)
        rte.init_absmdl('R20')
        df = rte.execute()

        df_expected = pd.read_csv(
            os.path.join(THIS_DIR, "data", "tb_tot_ros03_16_17_18_19_19sd_20_20sd_98_mak11_21sd.csv"))
        assert_allclose(df.tbtotal, df_expected.ros20, atol=0)

    # @pytest.mark.datafiles(DATA_DIR)
    @pytest.mark.skip(reason="skipping")
    def test_pyrtlib_sat_R20SD(self):
        z, p, _, t, md = atmp.gl_atm(atmp.TROPICAL)

        gkg = ppmv2gkg(md[:, atmp.H2O], atmp.H2O)
        rh = mr2rh(p, t, gkg)[0] / 100

        ang = np.array([90.])
        frq = np.arange(20, 201, 1)

        rte = TbCloudRTE(z, p, t, rh, frq, ang)
        rte.init_absmdl('R20SD')
        df = rte.execute()

        df_expected = pd.read_csv(
            os.path.join(THIS_DIR, "data", "tb_tot_ros03_16_17_18_19_19sd_20_20sd_98_mak11_21sd.csv"))
        assert_allclose(df.tbtotal, df_expected.ros20sd, atol=0)

    # @pytest.mark.datafiles(DATA_DIR)
    @pytest.mark.skip(reason="skipping")
    def test_pyrtlib_sat_R21SD(self):
        z, p, _, t, md = atmp.gl_atm(atmp.TROPICAL)

        gkg = ppmv2gkg(md[:, atmp.H2O], atmp.H2O)
        rh = mr2rh(p, t, gkg)[0] / 100

        frq = np.arange(20, 201, 1)

        rte = TbCloudRTE(z, p, t, rh, frq)
        rte.init_absmdl('R20')
        H2OAbsModel.model = 'R21SD'
        H2OAbsModel.set_ll()
        df = rte.execute()

        df_expected = pd.read_csv(
            os.path.join(THIS_DIR, "data", "tb_tot_ros03_16_17_18_19_19sd_20_20sd_98_mak11_21sd.csv"))
        assert_allclose(df.tbtotal, df_expected.ros21sd, atol=0)

    def test_pyrtlib_sat_R18(self):
        z, p, _, t, md = atmp.gl_atm(atmp.TROPICAL)

        gkg = ppmv2gkg(md[:, atmp.H2O], atmp.H2O)
        rh = mr2rh(p, t, gkg)[0] / 100

        ang = np.array([90.])
        frq = np.arange(20, 201, 1)

        rte = TbCloudRTE(z, p, t, rh, frq, ang)
        rte.init_absmdl('R18')
        df = rte.execute()

        df_expected = pd.read_csv(
            os.path.join(THIS_DIR, "data", "tb_tot_ros03_16_17_18_19_19sd_20_20sd_98_mak11_21sd.csv"))
        assert_allclose(df.tbtotal, df_expected.ros18, atol=0)

    def test_pyrtlib_sat_R98(self):
        z, p, _, t, md = atmp.gl_atm(atmp.TROPICAL)

        gkg = ppmv2gkg(md[:, atmp.H2O], atmp.H2O)
        rh = mr2rh(p, t, gkg)[0] / 100

        ang = np.array([90.])
        frq = np.arange(20, 201, 1)

        rte = TbCloudRTE(z, p, t, rh, frq, ang)
        rte.init_absmdl('R98')
        df = rte.execute()

        df_expected = pd.read_csv(
            os.path.join(THIS_DIR, "data", "tb_tot_ros03_16_17_18_19_19sd_20_20sd_98_mak11_21sd.csv"))
        assert_allclose(df.tbtotal, df_expected.rosen, atol=0)

    # @pytest.mark.datafiles(DATA_DIR)
    def test_pyrtlib_sat_R98_cloudy(self):
        z, p, _, t, md = atmp.gl_atm(atmp.TROPICAL)

        gkg = ppmv2gkg(md[:, atmp.H2O], atmp.H2O)
        rh = mr2rh(p, t, gkg)[0] / 100

        ang = np.array([90.])
        frq = np.arange(20, 201, 1)

        denliq = np.zeros(z.shape)
        denice = np.zeros(z.shape)
        cldh = np.zeros((2, 2))
        # build a cloud
        ib = 1
        it = 3
        denliq[ib:it + 1] = 10 * np.ones((it - ib + 1))
        cldh[:, 0] = np.array([z[ib], z[it]])
        ib = 29
        it = 31
        denice[ib:it + 1] = 0.1 * np.ones((it - ib + 1))
        cldh[:, 1] = np.array([z[ib], z[it]])

        rte = TbCloudRTE(z, p, t, rh, frq, ang)
        rte.init_absmdl('R98')
        rte.cloudy = True
        rte.init_cloudy(cldh, denice, denliq)
        df = rte.execute()

        df_expected = pd.read_csv(os.path.join(THIS_DIR, "data", "cloudy_tb_tot_ros98.csv"))
        assert_allclose(df.tbtotal, df_expected.ros98, atol=0)

    def test_pyrtlib_sat_no_raytracing(self):
        z, p, _, t, md = atmp.gl_atm(atmp.TROPICAL)

        gkg = ppmv2gkg(md[:, atmp.H2O], atmp.H2O)
        rh = mr2rh(p, t, gkg)[0] / 100

        ang = np.array([90.])
        frq = np.arange(20, 201, 1)

        rte = TbCloudRTE(z, p, t, rh, frq, ang)
        rte.ray_tracing = False
        rte.init_absmdl('R19SD')

        df = rte.execute()
        df_expected = pd.read_csv(
            os.path.join(THIS_DIR, "data", "tb_tot_ros03_16_17_18_19_19sd_20_20sd_98_mak11_21sd.csv"))
        assert_allclose(df.tbtotal, df_expected.ros19sd, atol=0)

    def test_pyrtlib_sat_rose21_wO3(self):
        z, p, _, t, md = atmp.gl_atm(atmp.US_STANDARD)

        o3n_ppmv = md[:, atmp.O3]
        o3n = np.zeros(z.shape)
        for k in range(0, len(z)):
            o3n[k] = ppmv_to_moleculesm3(o3n_ppmv[k], p[k] * 100.0, t[k])

        gkg = ppmv2gkg(md[:, atmp.H2O], atmp.H2O)
        rh = mr2rh(p, t, gkg)[0] / 100

        ang = np.array([90.])
        frq = np.arange(20, 201, 1)

        rte = TbCloudRTE(z, p, t, rh, frq, ang, o3n)
        rte.init_absmdl('R20')
        H2OAbsModel.model = 'R21SD'
        H2OAbsModel.set_ll()
        O3AbsModel.model = 'R18'
        O3AbsModel.set_ll()
        df = rte.execute()

        df_expected = pd.read_csv(
            os.path.join(THIS_DIR, "data", "tb_tot_ros21sd_wo3.csv"))
        assert_allclose(df.tbtotal, df_expected.ros21sd_wo3, atol=0)

    def test_pyrtlib_sat_R21SD_wyoming_es(self):
        date = datetime(2021, 4, 22, 12)
        station = 'LIRE'
        df_w = WyomingUpperAir.request_data(date, station)

        z, p, t, gkg = df_w.height.values / 1000, \
                    df_w.pressure.values, \
                    to_kelvin(df_w.temperature.values), \
                    df_w.mixr.values

        rh = dewpoint2rh(df_w.dewpoint, df_w.temperature).values

        ang = np.array([90.])
        frq = np.arange(20, 201, 1)

        rte = TbCloudRTE(z, p, t, rh, frq, ang)
        rte.emissivity = 0.6
        rte.init_absmdl('R20')
        H2OAbsModel.model = 'R21SD'
        H2OAbsModel.set_ll()
        df = rte.execute()

        df_expected = pd.read_csv(
            os.path.join(THIS_DIR, "data", "tb_tot_rose21sd_RAOB_es.csv"))
        assert_allclose(df.tbtotal, df_expected.tbtotal_wyoming, atol=0)
        
    def test_pyrtlib_sat_R21SD_igra2_es(self):
        date = datetime(2020, 6, 1, 12)
        station = 'SPM00008221'
        df_igra2, header = IGRAUpperAir.request_data(date, station)
        
        df_igra2 = df_igra2[df_igra2.pressure.notna() & 
                            df_igra2.temperature.notna() & 
                            df_igra2.dewpoint.notna() & 
                            df_igra2.height.notna()]

        z, p, t = df_igra2.height.values / 1000, df_igra2.pressure.values, to_kelvin(df_igra2.temperature.values)
        
        rh = dewpoint2rh(df_igra2.dewpoint, df_igra2.temperature).values

        ang = np.array([90.])
        frq = np.arange(20, 201, 1)

        rte = TbCloudRTE(z, p, t, rh, frq, ang)
        rte.emissivity = 0.6
        rte.init_absmdl('R20')
        H2OAbsModel.model = 'R21SD'
        H2OAbsModel.set_ll()
        df = rte.execute()

        df_expected = pd.read_csv(
            os.path.join(THIS_DIR, "data", "tb_tot_rose21sd_RAOB_es.csv"))
        assert_allclose(df.tbtotal, df_expected.tbtotal_igra2, atol=0)
        
    def test_pyrtlib_sat_R21SD_igra2_beg2021_es(self):
        date = datetime(2021, 10, 2, 0)
        station = 'ASM00094610'
        df_igra2, header = IGRAUpperAir.request_data(date, station, beg2021=True)
        
        df_igra2 = df_igra2[df_igra2.pressure.notna() & 
                            df_igra2.temperature.notna() & 
                            df_igra2.dewpoint.notna() & 
                            df_igra2.height.notna()]

        z, p, t = df_igra2.height.values / 1000, df_igra2.pressure.values, to_kelvin(df_igra2.temperature.values)
        
        rh = dewpoint2rh(df_igra2.dewpoint, df_igra2.temperature).values

        ang = np.array([90.])
        frq = np.arange(20, 201, 1)

        rte = TbCloudRTE(z, p, t, rh, frq, ang)
        rte.emissivity = 0.6
        rte.init_absmdl('R20')
        H2OAbsModel.model = 'R21SD'
        H2OAbsModel.set_ll()
        df = rte.execute()

        df_expected = pd.read_csv(
            os.path.join(THIS_DIR, "data", "tb_tot_rose21sd_RAOB_es.csv"))
        assert_allclose(df.tbtotal, df_expected.tbtotal_igra2_beg2021, atol=0)

    def test_pyrtlib_sat_R21SD_ERA5_cloudy(self):
        lonlat = (15.8158, 38.2663)
        nc_file = os.path.join(THIS_DIR, "data", "era5_reanalysis-2019-06-25T12:00:00.nc")
        df_era5 = ERA5Reanalysis.read_data(nc_file, lonlat)

        ang = np.array([90.])
        frq = np.arange(20, 201, 1)

        cldh = np.empty((2, 1))
        cldh[:, 0] = np.array([np.min(df_era5.z), np.max(df_era5.z)])

        total_mass = 1 - df_era5.ciwc.values - df_era5.clwc.values - df_era5.crwc.values - df_era5.cswc.values
        denice = df_era5.ciwc.values * (1/total_mass) * kgkg_to_kgm3(df_era5.q.values * (1/total_mass),
                                                    df_era5.p.values, df_era5.t.values) * 1000
        denliq = df_era5.clwc.values * (1/total_mass) * kgkg_to_kgm3(df_era5.q.values * (1/total_mass),
                                                    df_era5.p.values, df_era5.t.values) * 1000


        rte = TbCloudRTE(df_era5.z.values, df_era5.p.values, df_era5.t.values, df_era5.rh.values, frq, ang)
        rte.init_absmdl('R20')
        rte.init_cloudy(cldh, denice, denliq)
        H2OAbsModel.model = 'R21SD'
        H2OAbsModel.set_ll()
    
        df = rte.execute()

        df_expected = pd.read_csv(
            os.path.join(THIS_DIR, "data", "tb_tot_rose21sd_RAOB_es.csv"))
        assert_allclose(df.tbtotal, df_expected.tbtotal, atol=0)
        


import numpy as np
import os

PATH = os.path.dirname(os.path.abspath(__file__))
mtx = np.loadtxt(os.path.join(PATH, "h2o_list_r21.txt"))
# continuum terms
ctr = np.array([300.0, 5.919e-10, 3.0, 1.416e-08, 7.5])

# below is from abh2o_sd.f
# read line parameters; units: ghz, hz*cm^2, mhz/mb
reftline = 296.0
fl = mtx[:, 1]
s1 = mtx[:, 2]
b2 = mtx[:, 3]
w0 = mtx[:, 4] / 1000.0
x = mtx[:, 5]
w0s = mtx[:, 6] / 1000.0
xs = mtx[:, 7]
sh = mtx[:, 8] / 1000.0
xh = mtx[:, 9]
shs = mtx[:, 10] / 1000.0
xhs = mtx[:, 11]
aair = mtx[:, 12]
aself = mtx[:, 13]
w2 = mtx[:, 14] / 1000.0
xw2 = mtx[:, 15]
w2s = mtx[:, 16] / 1000.0
xw2s = mtx[:, 17]
d2 = mtx[:, 18] / 1000.0
d2s = mtx[:, 19] / 1000.0

reftcon = ctr[0]
cf = ctr[1]
xcf = ctr[2]
cs = ctr[3]
xcs = ctr[4]



# This module contains uncertainty covariance matrix for oxygen and water vapor parameters
# as estimated in Cimini et al. ACP 2018 (https://doi.org/10.5194/acp-18-15231-2018) and
# updated in Cimini et al. GMD 2019 (https://doi.org/10.5194/gmd-12-1833-2019).
# The covariance matrices were derived with the routines by Phil Rosenkranz (https://doi.org/10.21982/M81013),
# as downloaded on 15 June 2017, available at http://cetemps.aquila.infn.it/mwrnet/lblmrt_ns.html
#
# This version is VALID for ground-based RT calculations in the range 20-150 GHz.
#
# The list of 112 parameters with units is as follows:
#   1      : O2 S(300) [%]^2
#   2      : O2 n_a [adim]^2
#   3      : O2 gamma_0(300) [GHz/bar]^2
#   4- 37  : O2 gamma_a(300) [GHz/bar]^2 for N=1-,1+,3-,...,33-,33+ lines
#  38- 71  : O2 y(300) [1/bar]^2 for N=1-,1+,3-,...,33-,33+ lines
#  72-105  : O2 v [1/bar]^2 for N=1-,1+,3-,...,33-,33+ lines
# 106      : H2O C_f(300) [km-1 mb-2 GHz-2]^2
# 107      : H2O C_s(300) [km-1 mb-2 GHz-2]^2
# 108      : H2O gamma_a(296) at 22.2 GHz [GHz/bar]^2
# 109      : H2O S(296) at 22.2 GHz [Hz*cm2]^2
# 110      : H2O n_Cf [adim]^2
# 111      : H2O R at 22.2 GHz [adim]^2
# 112      : H2O n_Cs [adim]^2

import os

import numpy as np
from netCDF4 import Dataset

PATH = os.path.dirname(os.path.abspath(__file__))

nc = Dataset(os.path.join(PATH, "R17", "Cov_parameters_Cimini_et_al_2018_V1.1.nc"))
R17_111 = np.asarray(nc.variables['Cov_p'][:])

nc = Dataset(os.path.join(PATH, "R17", "Cov_parameters_Cimini_et_al_2019_V2.0.nc"))
R17_112 = np.asarray(nc.variables['Cov_p'][:])



import os
# from pathlib import Path
from unittest import TestCase

import numpy as np
import pandas as pd
from numpy.testing import assert_allclose, assert_equal
from pyrtlib.climatology import AtmosphericProfiles as atmp
from pyrtlib.tb_spectrum import TbCloudRTE
from pyrtlib.absorption_model import H2OAbsModel
from pyrtlib.apiwebservices import ERA5Reanalysis
from pyrtlib.utils import ppmv2gkg, mr2rh, import_lineshape

# TEST_DIR = Path(__file__).parent
# DATA_DIR = os.path.join(TEST_DIR, 'data')
THIS_DIR = os.path.dirname(os.path.abspath(__file__))


class Test(TestCase):

    # @pytest.mark.datafiles(DATA_DIR)
    def test_pyrtlib_ground_R19SD(self):
        z, p, _, t, md = atmp.gl_atm(atmp.TROPICAL)

        gkg = ppmv2gkg(md[:, atmp.H2O], atmp.H2O)
        rh = mr2rh(p, t, gkg)[0] / 100

        ang = np.array([90.])
        frq = np.arange(20, 201, 1)

        rte = TbCloudRTE(z, p, t, rh, frq, ang)
        rte.satellite = False
        rte.init_absmdl('R19SD')
        df = rte.execute()

        df_expected = pd.read_csv(os.path.join(THIS_DIR, "data", "tb_tot_ground_ros03_19sd_21sd_era5.csv"))
        assert_allclose(df.tbtotal, df_expected.ros19sd, atol=0)

    # @pytest.mark.datafiles(DATA_DIR)
    # @pytest.mark.skip(reason="R03 not completly implemented yet")
    def test_pyrtlib_ground_R03(self):
        z, p, _, t, md = atmp.gl_atm(atmp.TROPICAL)

        gkg = ppmv2gkg(md[:, atmp.H2O], atmp.H2O)
        rh = mr2rh(p, t, gkg)[0] / 100

        frq = np.arange(20, 201, 1)

        rte = TbCloudRTE(z, p, t, rh, frq)
        rte.satellite = False
        rte.init_absmdl('R03')
        df = rte.execute()

        df_expected = pd.read_csv(os.path.join(THIS_DIR, "data", "tb_tot_ground_ros03_19sd_21sd_era5.csv"))
        assert_allclose(df.tbtotal.values, df_expected.ros03)

    def test_pyrtlib_ground_R21SD_ERA5(self):
        lonlat = (15.8158, 38.2663)
        nc_file = os.path.join(THIS_DIR, "data", "era5_reanalysis-2019-06-25T12:00:00.nc")
        df_era5 = ERA5Reanalysis.read_data(nc_file, lonlat)

        ang = np.array([90.])
        frq = np.arange(20, 201, 1)

        rte = TbCloudRTE(df_era5.z.values, df_era5.p.values, df_era5.t.values, df_era5.rh.values, frq, ang)
        rte.satellite = False
        rte.init_absmdl('R20')
        H2OAbsModel.model = 'R21SD'
        H2OAbsModel.set_ll()
        df = rte.execute()

        df_expected = pd.read_csv(os.path.join(THIS_DIR, "data", "tb_tot_ground_ros03_19sd_21sd_era5.csv"))
        assert_allclose(df.tbtotal, df_expected.rose21sd_era5, atol=0)



from .erafive import ERA5Reanalysis
from .wyomingupperair import WyomingUpperAir
from .igra2 import IGRAUpperAir
# from .eps_sandbox import EUMETSATProduct

__all__ = ['WyomingUpperAir', 'IGRAUpperAir', 'ERA5Reanalysis']



# -*- coding: utf-8 -*-

"""REFERENCES FOR MEASUREMENTS (freq in GHz, S in Hz*cm^2, W's,D's in GHZ/bar, X's & A dimensionless) updated Dec. 30, 2020.

    References
    ----------
    .. [1] M. Koshelev et al., JQSRT v.205, pp. 51-58 (2018)
    .. [2] V. Payne et al.,IEEE Trans. Geosci. Rem. Sens. v.46, pp.3601-3617 (2008)
    .. [3] G. Golubiatnikov, J. MOLEC. SPEC. vol. 230, pp.196-198 (2005)
    .. [4] M. Koshelev et al., J. Molec. Spec. v.241, pp.101-108 (2007)
    .. [5] J.-M. Colmont et al.,J. Molec. Spec. v.193, pp.233-243 (1999)
    .. [6] M. Tretyakov et al, JQSRT v.114 pp.109-121 (2013)
    .. [7] G. Golubiatnikov et al., JQSRT v.109, pp.1828-1833 (2008)
    .. [8] V. Podobedov et al., JQSRT v.87, pp. 377-385 (2004)
    .. [9] M. Koshelev, JQSRT v.112, pp.550-552 (2011)
    .. [10] M. Tretyakov, JQSRT v.328, pp.7-26 (2016)
    .. [11] D. Turner et al., IEEE Trans. Geosci. Rem. Sens. v.47 pp.3326-37 (2009), re-adjusted for new line par. Aug.22, 2022.
    .. [11] M. Koshelev et al. JQSRT doi:10.1016/j.jqsrt.2020.107472

    Other parameters from HITRAN2020.
"""
# code to import coefficient from .asc file
# import pandas as pd
# import numpy as np

# a = pd.read_table("h2o_sdlist.asc", sep=',',header=None, skiprows=1, nrows=19, usecols=range(0, 20))
# a = np.loadtxt("h2o_sdlist.asc", delimiter=',', skiprows=1, usecols=range(0, 20), max_rows=20)
# np.savetxt("/Users/slarosa/dev/pyrtlib/pyrtlib/lineshape/h2o_list_r22.txt", a)

import numpy as np
import os

PATH = os.path.dirname(os.path.abspath(__file__))
mtx = np.loadtxt(os.path.join(PATH, "h2o_list_r22.txt"))

ctr = np.array([300.0, 5.919e-10, 3.0, 1.416e-08, 7.5])

# # below is from abh2o_sd.f
# # read line parameters; units: ghz, hz*cm^2, mhz/mb
reftline = 296.0
fl = mtx[:, 1]
s1 = mtx[:, 2]
b2 = mtx[:, 3]
w0 = mtx[:, 4] / 1000.0
x = mtx[:, 5]
w0s = mtx[:, 6] / 1000.0
xs = mtx[:, 7]
sh = mtx[:, 8] / 1000.0
xh = mtx[:, 9]
shs = mtx[:, 10] / 1000.0
xhs = mtx[:, 11]
aair = mtx[:, 12]
aself = mtx[:, 13]
w2 = mtx[:, 14] / 1000.0
xw2 = mtx[:, 15]
w2s = mtx[:, 16] / 1000.0
xw2s = mtx[:, 17]
d2 = mtx[:, 18] / 1000.0
d2s = mtx[:, 19] / 1000.0

reftcon = ctr[0]
cf = ctr[1]
xcf = ctr[2]
cs = ctr[3]
xcs = ctr[4]




"""
Read and Download data from EPS Datastore by EUMETSAT.
To download products from EUMETSAT it is necessary to configure a API key.
To get cinsumer_key and consumer_secret you have to registered to EPS system.
Credentials may be found to the following link https://api.eumetsat.int/api-key/
"""

__author__ = ''
__date__ = 'March 2023'
__copyright__ = '(C) 2021, CNR-IMAA'

import os
from pathlib import Path
import math
import warnings
from datetime import datetime
from typing import Optional, Tuple, List, Any

try:
    import eumdac
except ModuleNotFoundError as e:
    warnings.warn("Module EUMDAC must be installed to download EPS dataset.")
import numpy as np
# from scipy.spatial import cKDTree
from sklearn.neighbors import BallTree, KDTree
import pandas as pd
from netCDF4 import Dataset

from ..utils import pressure_to_height


class EUMETSATProduct:
    """Read and Download product from EUMETSAT Data Store Web Services"""
    def __init__(self) -> None:
        super().__init__()
        self.token = None
        self.datastore = None
        self.datatailoe = None
        self.tailor_models = None

    def authentication(self) -> None:
        rc_file = os.path.join(Path.home(), '.eumdacrc')
        if not os.path.exists(rc_file):
            return warnings.warn("eumdacrc file not found in home directory.")
        with open(rc_file, 'r') as f:
            lines = f.readlines()
            consumer_key = lines[0].split(":")[1].strip()
            consumer_secret = lines[1].split(":")[1].strip()
    
        credentials = (consumer_key, consumer_secret)
        self.token = eumdac.AccessToken(credentials)
        self.datastore = eumdac.DataStore(self.token)
        self.datatailoe = eumdac.DataTailor(self.token)
        self.tailor_models = eumdac.tailor_models
        print(f"This token '{self.token}' expires {self.token.expiration}")

    def list_collections(self, keyword: str=None) -> Tuple:
        """_summary_

        Args:
            keyword (str, optional): _description_. Defaults to None.

        Returns:
            Tuple: _description_
        """
        collections = self.datastore.collections
        ids = []
        if keyword:
            collections_filtered = []
            for cl in collections:
                if keyword in cl.title:
                    ids.append(cl.metadata['properties']['identifier'])
                    collections_filtered.append(cl)
            collections = collections_filtered
        else:
            for cl in collections:
                ids.append(cl.metadata['properties']['identifier'])
        
        return (ids, collections)
    
    def product_search(self, collection: eumdac.collection.Collection, **query: Any) -> eumdac.collection.SearchResults:
        """_summary_

        Args:
            collection (eumdac.collection.Collection): _description_

        Returns:
            eumdac.collection.SearchResults: _description_
        """
        products = collection.search(**query)
        
        return products

    def custom_chain(self):
        """_summary_

        Returns:
            _type_: _description_
        """
        pass

    @classmethod
    def read_data(cls, file: str, lonlat: tuple) -> pd.DataFrame:
        """Read data from the ERA5 Reanalysis dataset.

        Args:
            file (str): The netcdf file
            lonlat (tuple): longitude and latitude

        Returns:
            pandas.DataFrame: [description]

        .. note:: To convert specific cloud water content (CLWC) or specific cloud ice water content (CIWC) 
            from kg kg-1 to g m-3 using this function :py:meth:`pyrtlib.utils.kgkg_to_gm3`
        """
        ERAFIVE = cls()
        nc = Dataset(file)
        lats = nc.variables['latitude'][:]
        lons = nc.variables['longitude'][:]
        # idx_lat = ERAFIVE.find_nearest(lats, lonlat[1])
        # idx_lon = ERAFIVE.find_nearest(lons, lonlat[0])
        idx, dist = ERAFIVE.find_nearest(lons, lats, lonlat)

        pres = np.asarray(nc.variables['level'][:])
        temp = np.asarray(nc.variables['t'][:, :, idx, idx])
        rh = np.asarray(nc.variables['r'][:, :, idx, idx]) / 100 # RH in decimal
        clwc = np.asarray(nc.variables['clwc'][:, :, idx, idx])
        ciwc = np.asarray(nc.variables['ciwc'][:, :, idx, idx])
        crwc = np.asarray(nc.variables['crwc'][:, :, idx, idx])
        cswc = np.asarray(nc.variables['cswc'][:, :, idx, idx])
        q = np.asarray(nc.variables['q'][:, :, idx, idx])

        z = pressure_to_height(pres) / 1000 # Altitude in km
        date = pd.to_datetime(nc.variables['time'][:], origin='1900-01-01 00:00:00.0', unit='h')

        df = pd.DataFrame({'p': np.flip(pres),
                           'z': np.flip(z),
                           't': np.flip(temp[0]),
                           'rh': np.flip(rh[0]),
                           'clwc': np.flip(clwc[0]),
                           'ciwc': np.flip(ciwc[0]),
                           'crwc': np.flip(crwc[0]),
                           'cswc': np.flip(cswc[0]),
                           'q': np.flip(q[0]),
                           'time': np.repeat(date, len(z))
                           })

        return df #, (idx, dist), np.stack((lons, lats))

    def find_nearest(self, lons, lats, point):
        """Find index of nearest coordinate

        Args:
            array ([type]): [description]
            value ([type]): [description]

        Returns:
            [type]: [description]
        """
        # idx = np.searchsorted(array, value, side="left")
        # if np.amax(array) < value < np.amin(array):
        #     warnings.warn('Out of boundingbox: value {} is out of dataset extent'.format(value))
        # if idx > 0 and (idx == len(array) or math.fabs(value - array[idx - 1]) < math.fabs(value - array[idx])):
        #     return idx - 1
        # else:
        #     return idx

        # scipy
        # combined_x_y_arrays = np.dstack([lats.ravel(), lons.ravel()])[0]
        # # points = list(point.transpose())
        #
        # mytree = cKDTree(combined_x_y_arrays)
        # dist, indexes = mytree.query(point)

        ball = BallTree(np.radians(np.stack((lats, lons), axis=-1)), metric='haversine')
        dist, indexes = ball.query(np.radians(np.array(np.flip(point)).reshape(1, 2)), k=1)
        
        return indexes[0][0], dist

    @staticmethod
    def request_data(path: str, time: datetime, lonlat: tuple, resolution: Optional[float] = 0.25, offset: Optional[float] = 0.4) -> str:
        """Download ERA5Reanalysis data from the Copernicus Climate Change Service.

        Args:
            path (str): The output directory
            time (datetime): The date and time of the desired observation.
            lonlat (tuple): The coordinatre in degrees, longitude and latitude
            resolution (Optional[float], optional): The pixel size of the requested grid data. Defaults to 0.25.
            offset (Optional[float], optional): The offset to apply to coordinates to get the extent. Defaults to 0.3.

        Returns:
            str: The path to downloaded netcdf file
        """
        # North, West, South, Est
        extent = [lonlat[1] + offset, lonlat[0] - offset, lonlat[1] - offset, lonlat[0] + offset]
        nc_file_name = 'era5_reanalysis-{}.nc'.format(time.isoformat())
        nc_file = os.path.join(path, nc_file_name)

        variables = ['relative_humidity', 'specific_cloud_ice_water_content', 'specific_cloud_liquid_water_content',
                     'specific_humidity', 'specific_rain_water_content', 'specific_snow_water_content', 'temperature']
        return nc_file



from setuptools import setup, find_packages

with open('requirements.txt') as f:
    required = f.read().splitlines()

setup(
    name='pyrtlib',
    version='1.0.0',
    include_package_data=True,
    package_dir={'': 'pyrtlib'},
    python_requires='>=3.6',
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    install_requires=required,
    url='',
    license='GPLv3',
    author='slarosa',
    author_email='salvatore-larosa@cnr.it',
    description='pyrtlib - Radiative Transfer library',
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Education',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Topic :: Scientific/Engineering',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    packages=find_packages('pyrtlib')
)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from typing import Tuple, Optional
from datetime import datetime
import numpy as np

from pyrtlib.utils import rho2rh


class ProfileExtrapolation:

    # This is an abstract class that contains an instance to a modality of the
    # extrapolation.

    def __init__(self, mode: Optional[str] ='ITU-Annex1'):
        """Initialize what extrapolation to be used.
        Only ITU-835-6 Annex1 has been implemented rigth now.

        Args:
            mode (str, optional): Recommendation of extrapolation. Defaults to 'ITU-Annex1'.

        Raises:
            ValueError: Raise an error if mode is not set or if unsupported.

        References
        ----------
            .. [1] https://www.itu.int/rec/R-REC-P.835/en
        """
        if mode == 'ITU-Annex1':
            self.instance = _ITU835_6()
            self._height = np.hstack((np.linspace(0, 25, 26),
                                      np.linspace(27.5, 50, 10),
                                      np.linspace(55, 100, 10)))
        else:
            raise ValueError("Modality of extrapolation is mandatory")

    @property
    def height(self) -> np.ndarray:
        """Getter/Setter for altitude vector"""
        return self._height

    @height.setter
    def height(self, height: np.ndarray) -> None:
        """Setter for absorption model"""
        if height is not None and isinstance(height, np.ndarray):
            self._height = height
        else:
            raise ValueError("Please enter a valid array for height")

    def temperature(self, lat: float, h: np.ndarray, season: Optional[str] = 'summer') -> np.ndarray:
        """ Determine the temperature at a given latitude and height.

        Method to determine the temperature as a function of altitude and latitude,
        for calculating gaseous attenuation along an Earth-space path. This method
        is recommended when more reliable local data are not available.


        Parameters
        ----------
        lat : numpy.ndarray
            Latitude (degree)
        h : numpy.ndarray
            Height (km)
        season : str
            Season of the year (available values, 'summer', and 'winter').
            Default 'summer'


        Returns
        -------
        T: numpy.ndarray
            Temperature (K)


        References
        ----------
        [1] Reference Standard Atmospheres
        https://www.itu.int/rec/R-REC-P.835/en

        """
        return self.instance.temperature(lat, h, season)

    def pressure(self, lat: float, h: np.ndarray, season: Optional[str] = 'summer') -> np.ndarray:
        """ Determine the atmospheric pressure at a given latitude and height.

        Method to determine the pressure as a function of altitude and latitude,
        for calculating gaseous attenuation along an Earth-space path.
        This method is recommended when more reliable local data are not available.

        Parameters
        ----------
        lat : numpy.ndarray
            Latitude (degree)
        h : numpy.ndarray
            Height (km)
        season : str
            Season of the year (available values, 'summer', and 'winter').
            Default 'summer'


        Returns
        -------
        P: numpy.ndarray
            Pressure (hPa)


        References
        ----------
        [1] Reference Standard Atmospheres
        https://www.itu.int/rec/R-REC-P.835/en
        """
        return self.instance.pressure(lat, h, season)

    def water_vapour_density(self, lat: float, h: np.ndarray, season: Optional[str] = 'summer') -> np.ndarray:
        """ Determine the water vapour density at a given latitude and height.

        Method to determine the water-vapour density as a
        function of altitude and latitude, for calculating gaseous attenuation
        along an Earth-space path. This method is recommended when more reliable
        local data are not available.


        Parameters
        ----------
        lat : numpy.ndarray
            Latitude (degree)
        h : numpy.ndarray
            Height (km)
        season : str
            Season of the year (available values, 'summer', and 'winter').
            Default 'summer'


        Returns
        -------
        rho: numpy.ndarray
            Water vapour density (g/m^3)


        References
        ----------
        [1] Reference Standard Atmospheres
        https://www.itu.int/rec/R-REC-P.835/en
        """
        return self.instance.water_vapour_density(lat, h, season)

    def standard_temperature(self, h: np.ndarray, T_0: float) -> np.ndarray:
        """ Determine the standard temperature at a given height.

        Method to compute the temperature of an standard atmosphere at
        a given height. The reference standard atmosphere is based on the United
        States Standard Atmosphere, 1976, in which the atmosphere is divided into
        seven successive layers showing linear variation with temperature.


        Parameters
        ----------
        h : numpy.ndarray
            Height (km)
        T_0 : float
            Surface temperature (K)


        Returns
        -------
        T: numpy.ndarray
            Temperature (K)


        References
        ----------
        [1] Reference Standard Atmospheres
        https://www.itu.int/rec/R-REC-P.835/en
        """
        return self.instance.standard_temperature(h, T_0)

    def standard_pressure(self, h: np.ndarray, T_0: float, P_0: float) -> np.ndarray:
        """ Determine the standard pressure at a given height.

        Method to compute the total atmopsheric pressure of an standard atmosphere
        at a given height.

        The reference standard atmosphere is based on the United States Standard
        Atmosphere, 1976, in which the atmosphere is divided into seven successive
        layers showing linear variation with temperature.


        Parameters
        ----------
        h : numpy.ndarray
            Height (km)
        T_0 : float
            Surface temperature (K)
        P_0 : float
            Surface pressure (hPa)


        Returns
        -------
        P: numpy.ndarray
            Pressure (hPa)


        References
        ----------
        [1] Reference Standard Atmospheres
        https://www.itu.int/rec/R-REC-P.835/en
        """
        return self.instance.standard_pressure(h, T_0, P_0)

    def standard_water_vapour_density(self, h: np.ndarray, h_0: float, rho_0: float) -> np.ndarray:
        """ Determine the standard water vapour density at a given height.

        The reference standard atmosphere is based on the United States Standard
        Atmosphere, 1976, in which the atmosphere is divided into seven successive
        layers showing linear variation with temperature.


        Parameters
        ----------
        h : numpy.ndarray
            Height (km)
        h_0 : float
            Scale height (km)
        rho_0 : float
            Surface water vapour density (g/m^3)


        Returns
        -------
        rho: numpy.ndarray
            Water vapour density (g/m^3)


        References
        ----------
        [1] Reference Standard Atmospheres
        https://www.itu.int/rec/R-REC-P.835/en
        """
        return self.instance.standard_water_vapour_density(h, h_0, rho_0)

    def standard_water_vapour_pressure(self, h: np.ndarray, h_0: Optional[float] = 2., rho_0: Optional[float] = 7.5) -> np.ndarray:
        """Determine the standard water vapour pressure at a given height.

        The reference standard atmosphere is based on the United States Standard
        Atmosphere, 1976, in which the atmosphere is divided into seven successive
        layers showing linear variation with temperature.


        Parameters
        ----------
        h : numpy.ndarray
            Height (km)
        h_0 : float
            Scale height (km)
            Default 2.
        rho_0 : float
            Surface water vapour density (g/m^3)
            Default 7.5


        Returns
        -------
        e: numpy.ndarray
            Water vapour pressure (hPa)


        References
        ----------
        [1] Reference Standard Atmospheres
        https://www.itu.int/rec/R-REC-P.835/en
        """
        
        return self.instance.standard_water_vapour_pressure(h, h_0, rho_0)

    def _get_season(self, lat: float, month: int) -> str:
        """_summary_

        Args:
            lat (float): _description_
            month (int): _description_

        Returns:
            str: _description_
        """
        if (lat > 20 and month in [6, 7, 8]) or (lat < -20 and month in [1, 2, 12]):
            season = 'summer'
        elif (lat > 20 and month in [1, 2, 12]) or (lat < -20 and month in [6, 7, 8]):
            season = 'winter'
        else:
            season = 'standard'

        return season

    def profile_extrapolation(self, lat: float, month: int, z: np.ndarray, q: Tuple[np.ndarray, np.ndarray, np.ndarray]) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Extrapolation of atmospheres to be used to determine 
        temperature, pressure and water-vapour pressure as a function 
        of altitude and latitude, for calculating gaseous attenuation when more reliable 
        local data are not available.

        Args:
            lat (float): Latitude (degree)
            month (int): Month of the year
            z (np.ndarray): Height (km)
            q (Tuple[np.ndarray, np.ndarray, np.ndarray]): Pressure (hPa), Temperature (K) and Relative humisity (frac) profiles

        Returns:
            Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]: Height, Pressure, Temperature, RH extrapolated profiles
        """
        if np.max(z) < 50:
            h_km = np.append(z, np.arange(max(z)+5, 50, 5))
            idx = np.where(self._height > np.max(h_km))
            self._height = np.append(h_km, self._height[idx])
        else:
            self._height = np.append(z, self._height[idx])

        idx = np.where(self._height > np.max(z))
        season = self._get_season(lat, month)

        p, t, rh = q

        if season != 'standard':
            tt = self.instance.temperature(lat, self._height, season)
            pp = self.instance.pressure(lat, self._height, season)
            wvd = self.instance.water_vapour_density(lat, self._height, season)
        else:
            tt = self.instance.standard_temperature(self._height, season)
            pp = self.instance.standard_pressure(self._height, season)
            wvd = self.instance.standard_water_vapour_density(
                self._height, season)

        pres = np.append(p, pp[idx])
        temp = np.append(t, tt[idx])
        _rh = rho2rh(wvd, tt, pp)[0]
        rh = np.append(rh, _rh[idx])

        return self._height, pres, temp, rh


class _ITU835_6():
    """Class to model the ITU-R P.835_6 recommendation.

    The procedures to compute the reference standard atmosphere parameters
    pressented in these versions are identical to those included in version
    ITU_T P.835.

    Version:
       * P.835-6 (12/17) (Current version)
    """

    def __init__(self):
        self.__version__ = 6
        self.year = 2017
        self.month = 12
        self.link = 'https://www.itu.int/rec/R-REC-P.835-6-201712-I/en'
        self.h_km = np.hstack((np.linspace(0, 25, 26),  np.linspace(27.5, 50, 10),
                               np.linspace(55, 100, 10)))

    @staticmethod
    def standard_temperature(h, T_0=288.15):
        """

        """
        h_p = 6356.766 * h / (6356.766 + h)
        # Warnings because of sqrt are expected
        with np.errstate(invalid='ignore'):
            T = np.where(h_p <= 11, 288.15 - 6.5 * h_p,
                         np.where(np.logical_and(11 < h_p, h_p <= 20),
                                  216.65,
                                  np.where(np.logical_and(20 < h_p, h_p <= 32),
                                  216.65 + (h_p - 20),
                                  np.where(np.logical_and(32 < h_p, h_p <= 47),
                                           228.65 + 2.8 * (h_p - 32),
                                           np.where(np.logical_and(47 < h_p, h_p <= 51),
                                           270.65,
                                           np.where(np.logical_and(51 < h_p, h_p <= 71),
                                                    270.65 - 2.8 * (h_p - 51),
                                                    np.where(np.logical_and(71 < h_p, h_p <= 84.852),
                                                    214.65 - 2.0 * (h_p - 71),
                                                    np.where(np.logical_and(86 <= h, h <= 91),
                                                             186.8673,
                                                             np.where(np.logical_and(91 < h, h <= 100),
                                                             263.1905 - 76.3232 *
                                                             np.sqrt(
                                                                 (1 - ((h - 91)/19.9429)**2)),
                                                        195.08134)))))))))

        return T

    @staticmethod
    def standard_pressure(h, T_0=None, P_0=None):
        """

        """
        h_p = 6356.766 * h / (6356.766 + h)
        with np.errstate(invalid='ignore'):
            P = np.where(h_p <= 11,
                         1013.25 * (288.15 / (288.15 - 6.5 * h_p)
                                    )**(-34.1632 / 6.5),
                         np.where(np.logical_and(11 < h_p, h_p <= 20),
                                  226.3226 *
                                  np.exp(-34.1632 * (h_p - 11) / 216.65),
                                  np.where(np.logical_and(20 < h_p, h_p <= 32),
                                  54.74980 * (216.65 / (216.65 + (h_p - 20))
                                              ) ** 34.1632,
                                  np.where(np.logical_and(32 < h_p, h_p <= 47),
                                           8.680422 * (228.65 / (228.65 + 2.8 * (h_p - 32))) **
                                           (34.1632 / 2.8),
                                           np.where(np.logical_and(47 < h_p, h_p <= 51),
                                                    1.109106 *
                                                    np.exp(-34.1632 *
                                                           (h_p - 47) / 270.65),
                                                    np.where(np.logical_and(51 < h_p, h_p <= 71),
                                                             0.6694167 * (270.65 / (270.65 - 2.8 *
                                                                                    (h_p - 51)))**(-34.1632 / 2.8),
                                                             np.where(np.logical_and(71 < h_p, h_p <= 84.852),
                                                                      0.03956649 *
                                                                      (214.65 / (214.65 - 2.0 *
                                                                       (h_p - 71)))**(-34.1632 / 2.0),
                                                                      np.where(np.logical_and(86 <= h, h <= 100),
                                                                               np.exp(95.571899 - 4.011801 * h + 6.424731e-2 * h**2 -
                                                                                      4.789660e-4 * h**3 + 1.340543e-6 * h**4),
                                                                               1e-62)))))))).astype(float)

        return P

    @staticmethod
    def standard_water_vapour_density(h, h_0=2, rho_0=7.5):
        """

        """
        return rho_0 * np.exp(-h / h_0)

    def standard_water_vapour_pressure(self, h, h_0=2, rho_0=7.5):
        """

        """
        rho_h = self.standard_water_vapour_density(h, h_0, rho_0)
        T_h = self.standard_temperature(h)
        return rho_h * T_h / 216.7

    #  Low latitude standard atmosphere functions  (Section ITU-R P.835-6 2)  #
    @staticmethod
    def low_latitude_temperature(h):
        """Section 2 of Recommendation ITU-R P.835-6."""
        return np.where(np.logical_and((0 <= h), (h < 17)),
                        300.4222 - 6.3533 * h + 0.005886 * h**2,
                        np.where(np.logical_and((17 <= h), (h < 47)),
                        194 + (h - 17) * 2.533,
                        np.where(np.logical_and((47 <= h), (h < 52)), 270,
                                 np.where(np.logical_and((52 <= h), (h < 80)),
                                 270 - (h - 52) * 3.0714,
                            np.where(np.logical_and((80 <= h), (h <= 100)), 184, 184)))))

    def low_latitude_pressure(self, h):
        """Section 2 of Recommendation ITU-R P.835-6."""
        P10 = 284.8526   # Pressure at 10 km using the equation below
        P72 = 0.0313660  # Pressure at 72 km using the equation below
        return np.where(np.logical_and((0 <= h), (h <= 10)),
                        1012.0306 - 109.0338 * h + 3.6316 * h**2,
                        np.where(np.logical_and((10 < h), (h <= 72)),
                        P10 * np.exp(-0.147 * (h - 10)),
                        np.where(np.logical_and((72 < h), (h <= 100)),
                                 P72 * np.exp(-0.165 * (h - 72)), np.nan)))

    @staticmethod
    def low_latitude_water_vapour(h):
        """Section 3.1 of Recommendation ITU-R P.835-6."""
        return np.where(np.logical_and((0 <= h), (h <= 15)), 19.6542 *
                        np.exp(-0.2313 * h - 0.1122 * h**2 + 0.01351 * h**3 -
                               0.0005923 * h**4), 0)

    # Mid latitude standard atmosphere functions  (Section ITU-R P.835-6 3)
    @staticmethod
    def mid_latitude_temperature_summer(h):
        """Section 3.1 of Recommendation ITU-R P.835-6."""
        return np.where(np.logical_and((0 <= h), (h < 13)),
                        294.9838 - 5.2159 * h - 0.07109 * h**2,
                        np.where(np.logical_and((13 <= h), (h < 17)), 215.15,
                        np.where(np.logical_and((17 <= h), (h < 47)),
                                 215.15 * np.exp((h - 17) * 0.008128),
                                 np.where(np.logical_and((47 <= h), (h < 53)), 275,
                                 np.where(np.logical_and((53 <= h), (h < 80)),
                                          275 + 20 *
                                          (1 - np.exp((h - 53) * 0.06)),
                                          np.where(np.logical_and((80 <= h), (h <= 100)),
                                          175, np.nan))))))

    def mid_latitude_pressure_summer(self, h):
        """Section 3.1 of Recommendation ITU-R P.835-6."""
        P10 = 283.7096    # Pressure at 10 km using the equation below
        P72 = 0.03124022  # Pressure at 72 km using the equation below
        return np.where(
            np.logical_and((0 <= h), (h <= 10)),
            1012.8186 - 111.5569 * h + 3.8646 * h**2, np.where(
                np.logical_and((10 < h), (h <= 72)),
                P10 * np.exp(-0.147 * (h - 10)),
                np.where(
                    np.logical_and((72 < h), (h <= 100)),
                    P72 * np.exp(-0.165 * (h - 72)),
                    np.nan)))

    @staticmethod
    def mid_latitude_water_vapour_summer(h):
        """Section 3.1 of Recommendation ITU-R P.835-6."""
        return np.where(np.logical_and((0 <= h), (h <= 15)),
                        14.3542 * np.exp(-0.4174 * h - 0.02290 * h**2 +
                                         0.001007 * h**3), 0)

    @staticmethod
    def mid_latitude_temperature_winter(h):
        """Section 3.2 of Recommendation ITU-R P.835-6."""
        return np.where(np.logical_and((0 <= h), (h < 10)),
                        272.7241 - 3.6217 * h - 0.1759 * h**2,
                        np.where(np.logical_and((10 <= h), (h < 33)), 218,
                        np.where(np.logical_and((33 <= h), (h < 47)),
                                 218 + (h - 33) * 3.3571,
                                 np.where(np.logical_and((47 <= h), (h < 53)), 265,
                                 np.where(np.logical_and((53 <= h), (h < 80)),
                                          265 - (h - 53) * 2.0370,
                                          np.where(np.logical_and((80 <= h), (h <= 100)),
                                                   210, np.nan))))))

    def mid_latitude_pressure_winter(self, h):
        """Section 3.2 of Recommendation ITU-R P.835-6."""
        P10 = 258.9787    # Pressure at 10 km using the equation below
        P72 = 0.02851702  # Pressure at 72 km using the equation below
        return np.where(np.logical_and((0 <= h), (h <= 10)),
                        1018.8627 - 124.2954 * h + 4.8307 * h**2,
                        np.where(np.logical_and((10 < h), (h <= 72)),
                        P10 * np.exp(-0.147 * (h - 10)),
                        np.where(np.logical_and((72 < h), (h <= 100)),
                                 P72 * np.exp(-0.155 * (h - 72)), np.nan)))

    @staticmethod
    def mid_latitude_water_vapour_winter(h):
        """Section 3.2 of Recommendation ITU-R P.835-6."""
        return np.where(np.logical_and(0 <= h, h <= 10),
                        3.4742 * np.exp(- 0.2697 * h - 0.03604 * h**2 +
                                        0.0004489 * h**3), 0)

    #  High latitude standard atmosphere functions  (Section ITU-R P.835-6 4)  #
    @staticmethod
    def high_latitude_temperature_summer(h):
        """Section 4.1 of Recommendation ITU-R P.835-6."""
        return np.where(np.logical_and((0 <= h), (h < 10)),
                        286.8374 - 4.7805 * h - 0.1402 * h**2,
                        np.where(np.logical_and((10 <= h), (h < 23)), 225,
                        np.where(np.logical_and((23 <= h), (h < 48)),
                                 225 * np.exp((h - 23) * 0.008317),
                                 np.where(np.logical_and((48 <= h), (h < 53)), 277,
                                 np.where(np.logical_and((53 <= h), (h < 79)),
                                          277 - (h - 53) * 4.0769,
                                          np.where(np.logical_and((79 <= h), (h <= 100)),
                                                   171, np.nan))))))

    def high_latitude_pressure_summer(self, h):
        """Section 4.1 of Recommendation ITU-R P.835-6."""
        P10 = 269.6138    # Pressure at 10 km using the equation below
        P72 = 0.04582115  # Pressure at 72 km using the equation below
        return np.where(np.logical_and((0 <= h), (h <= 10)),
                        1008.0278 - 113.2494 * h + 3.9408 * h**2,
                        np.where(np.logical_and((10 < h), (h <= 72)),
                        P10 * np.exp(-0.140 * (h - 10)),
                        np.where(np.logical_and((72 < h), (h <= 100)),
                                 P72 * np.exp(-0.165 * (h - 72)), np.nan)))

    @staticmethod
    def high_latitude_water_vapour_summer(h):
        """Section 4.1 of Recommendation ITU-R P.835-6."""
        return np.where(np.logical_and((0 <= h), (h <= 15)),
                        8.988 * np.exp(- 0.3614 * h - 0.005402 * h**2 -
                                       0.001955 * h**3), 0)

    @staticmethod
    def high_latitude_temperature_winter(h):
        """Section 4.2 of Recommendation ITU-R P.835-6."""
        return np.where(np.logical_and((0 <= h), (h < 8.5)),
                        257.4345 + 2.3474 * h - 1.5479 * h**2 +
                        0.08473 * h**3,
                        np.where(np.logical_and((8.5 <= h), (h < 30)), 217.5,
                        np.where(np.logical_and((30 <= h), (h < 50)),
                                 217.5 + (h - 30) * 2.125,
                                 np.where(np.logical_and((50 <= h), (h < 54)), 260,
                                 np.where(np.logical_and((54 <= h), (h <= 100)),
                                          260 - (h - 54) * 1.667, np.nan)))))

    def high_latitude_pressure_winter(self, h):
        """Section 4.2 of Recommendation ITU-R P.835-6."""
        P10 = 243.8718    # Pressure at 10 km using the equation below
        P72 = 0.02685355  # Pressure at 72 km using the equation below
        return np.where(np.logical_and((0 <= h), (h <= 10)),
                        1010.8828 - 122.2411 * h + 4.554 * h**2,
                        np.where(np.logical_and((10 < h), (h <= 72)),
                        P10 * np.exp(-0.147 * (h - 10)),
                        np.where(np.logical_and((72 < h), (h <= 100)),
                                 P72 * np.exp(-0.150 * (h - 72)), np.nan)))

    @staticmethod
    def high_latitude_water_vapour_winter(h):
        """Section 4.2 of Recommendation ITU-R P.835-6."""
        return np.where(np.logical_and((0 <= h), (h <= 10)),
                        1.2319 * np.exp(0.07481 * h - 0.0981 * h**2 +
                                        0.00281 * h**3), 0)

    def temperature(self, lat, h, season='summer'):
        """ Section 2 of Recommendation ITU-R P.835-6."""
        if season == 'summer':
            return np.where(
                np.abs(lat) < 22, self.low_latitude_temperature(h),
                np.where(
                    np.abs(lat) < 45, self.mid_latitude_temperature_summer(h),
                    self.high_latitude_temperature_summer(h)))
        elif season == 'winter':
            return np.where(
                np.abs(lat) < 22, self.low_latitude_temperature(h),
                np.where(
                    np.abs(lat) < 45, self.mid_latitude_temperature_winter(h),
                    self.high_latitude_temperature_winter(h)))
        else:
            raise ValueError("The value for argument 'season' is not correct."
                             "Valid values are 'summer' and 'winter'.")

    def pressure(self, lat, h, season='summer'):
        """ Section 2 of Recommendation ITU-R P.835-6."""
        if season == 'summer':
            return np.where(
                np.abs(lat) < 22, self.low_latitude_pressure(h),
                np.where(
                    np.abs(lat) < 45, self.mid_latitude_pressure_summer(h),
                    self.high_latitude_pressure_summer(h)))
        elif season == 'winter':
            return np.where(
                np.abs(lat) < 22, self.low_latitude_pressure(h),
                np.where(
                    np.abs(lat) < 45, self.mid_latitude_pressure_winter(h),
                    self.high_latitude_pressure_winter(h)))
        else:
            raise ValueError("The value for argument 'season' is not correct."
                             "Valid values are 'summer' and 'winter'")

    def water_vapour_density(self, lat, h, season='summer'):
        """ Section 2 of Recommendation ITU-R P.835-6."""
        if season == 'summer':
            return np.where(
                np.abs(lat) < 22, self.low_latitude_water_vapour(h),
                np.where(
                    np.abs(lat) < 45, self.mid_latitude_water_vapour_summer(h),
                    self.high_latitude_water_vapour_summer(h)))
        elif season == 'winter':
            return np.where(
                np.abs(lat) < 22, self.low_latitude_water_vapour(h),
                np.where(
                    np.abs(lat) < 45, self.mid_latitude_water_vapour_winter(h),
                    self.high_latitude_water_vapour_winter(h)))
        else:
            raise ValueError("The value for argument 'season' is not correct."
                             "Valid values are 'summer' and 'winter'")

